# Copyright 2019 RoadrunnerWMC
#
# This file is part of ndspy.
#
# ndspy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# ndspy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with ndspy.  If not, see <https://www.gnu.org/licenses/>.
"""
Support for SSARs.
"""

import struct

from . import _common
from . import soundSequence


class SSAR:
    """
    A SSAR*sequence archive file. This contains a blob of sequence
    events data, and a list of "sequences" that are essentially just
    pointers to starting locations in that data.
    """
    # When saving SDAT, two otherwise identical SSARs will share data
    # only if their dataMergeOptimizationIDs are the same.
    # You can pretty safely ignore this.
    dataMergeOptimizationID = 0

    _events = None
    sequences = None

    def __init__(self, file=None, unk02=0, names=None):
        self.unk02 = unk02

        self.sequences = []

        if file is not None:
            if not file.startswith(b'SSAR'):
                raise ValueError("Wrong magic (should be b'SSAR', instead"
                                 f' found {file[:4]})')

            # Read SSAR header
            magic, bom, version, filesize, headersize, numblocks = \
                _common.NDS_STD_FILE_HEADER.unpack_from(file, 0)
            if version != 0x100:
                raise ValueError(f'Unsupported SSAR version: {version}')
            assert magic == b'SSAR'

            # Read DATA block header
            dataMagic, dataSize, dataOffset, dataCount = \
                struct.unpack_from('<4s3I', file, 0x10)
            assert dataMagic == b'DATA'

            # Pad the length of the names list to the number of
            # sequences
            names = list(names) if names is not None else []
            while len(names) < dataCount:
                names.append(None)

            self.eventsData = file[dataOffset:filesize]

            dataArrayPos = 0x20
            for i in range(dataCount):
                (sequenceOffset, bankID, volume, channelPressure,
                        polyphonicPressure, playerID) = \
                    struct.unpack_from('<iH4B', file, dataArrayPos)
                dataArrayPos += 0xC

                seq = SSARSequence(sequenceOffset, bankID, volume,
                    channelPressure, polyphonicPressure, playerID,
                    parsed=False)

                self.sequences.append((names[i], seq))

            self.parsed = False
        else:
            self._events = []
            self.parsed = True


    @classmethod
    def fromEventsAndSequences(cls, events, sequences, unk02=0):
        """
        Create a new SSAR object from a list of sequence events and a
        list of sequences.
        """
        obj = cls(unk02=unk02)
        obj.events = events
        obj.sequences = sequences
        return obj


    @classmethod
    def fromFile(cls, filePath, *args, **kwargs):
        """
        Load an SSAR from a filesystem file.
        """
        with open(filePath, 'rb') as f:
            return cls(f.read(), *args, **kwargs)


    @property
    def events(self):
        if not self.parsed:
            raise ValueError('You must parse the SSAR with .parse()'
                             ' before you can access .events!')
        return self._events
    @events.setter
    def events(self, value):
        self._events = value


    @property
    def eventsData(self):
        if self.parsed:
            raise ValueError('You cannot use .eventsData after you have'
                             ' parsed the SSAR!')
        return self._eventsData
    @eventsData.setter
    def eventsData(self, value):
        self._eventsData = value


    def parse(self):
        """
        Attempt to process .eventsData to create .events. If successful,
        this switches the SSAR from the unparsed to the parsed state.
        """
        if self.parsed: return
        self._initFromData(self._eventsData)
        self.parsed = True
        del self._eventsData


    def _initFromData(self, eventsData):
        """
        Finish initializing the SSAR using events data.
        """

        startOffs = []
        startOffs2Seq = {}
        for i, (seqName, seq) in enumerate(self.sequences):
            if seq.firstEventOffset not in [-1, None]:
                startOffs.append(seq.firstEventOffset)

                # Put lists in the dictionary, because there can be
                # multiple sequences referencing the same offset
                if seq.firstEventOffset not in startOffs2Seq:
                    startOffs2Seq[seq.firstEventOffset] = []
                startOffs2Seq[seq.firstEventOffset].append(seq)

        self.events, startEvents = soundSequence.readSequenceEvents(
            eventsData, startOffs)

        for event, originalOff in zip(startEvents, startOffs):
            for seq in startOffs2Seq[originalOff]:
                # These have to happen in this order
                seq.parsed = True
                seq.firstEvent = event


    def save(self):
        """
        Generate file data representing this SSAR, and then return that
        data, .unk02, and a list of sequence names as a triple. This
        matches the parameters of the default class constructor.
        """

        tableData = bytearray()

        if self.parsed:
            startEvents = []
            seq2StartEvent = {}
            for seqName, seq in self.sequences:
                if not seq.parsed:
                    raise ValueError('Attempting to save a parsed SSAR, but'
                                     f' {seqName} is not parsed!')

                if seq.firstEvent is not None:
                    startEvents.append(seq.firstEvent)
                    seq2StartEvent[seq] = seq.firstEvent
            eventsData, startOffs = soundSequence.saveSequenceEvents(self.events,
                                                                   startEvents)

            for seqName, seq in self.sequences:
                if seq in seq2StartEvent:
                    off = startOffs[startEvents.index(seq2StartEvent[seq])]
                else:
                    off = -1
                s = struct.pack('<iH4B2x', off, *(seq.save()[1:]))
                tableData.extend(s)

        else:
            eventsData = self.eventsData
            for seqName, seq in self.sequences:
                if seq.parsed:
                    raise ValueError('Attempting to save an unparsed SSAR, but'
                                     f' {seqName} is parsed!')

                feo = seq.firstEventOffset
                if feo is None: feo = -1
                s = struct.pack('<iH4B2x', feo, *(seq.save()[1:]))
                tableData.extend(s)

        dataOffset = 0x20 + len(self.sequences) * 0xC
        fileLen = dataOffset + len(eventsData)

        data = bytearray()
        data.extend(_common.NDS_STD_FILE_HEADER.pack(b'SSAR', 0xFEFF, 0x100, fileLen, 16, 1))

        data.extend(struct.pack('<4s3I',
            b'DATA', fileLen - 0x10, dataOffset, len(self.sequences)))

        data.extend(tableData)
        data.extend(eventsData)

        return (data,
                self.unk02,
                [seqName for (seqName, seq) in self.sequences])


    def saveToFile(self, filePath):
        """
        Generate file data representing this SSAR, and save it to a
        filesystem file.
        """
        d = self.save()[0]
        with open(filePath, 'wb') as f:
            f.write(d)


    def __str__(self):
        if not self.parsed:
            names = [str(n) for (n, s) in self.sequences]
            return f'<ssar unparsed [{", ".join(names)}]>'

        linesList = ['<ssar']
        linesList.append(soundSequence.printSequenceEventList(
            self.events,
            {seqName: seq.firstEvent for (seqName, seq) in self.sequences},
            ' ' * 4))
        linesList.append('>')
        return '\n'.join(linesList)


    def __repr__(self):
        if self.parsed:
            return f'{type(self).__name__}.fromEventsAndSequences({self.events!r}, {self.sequences!r})'

        else:
            data = _common.shortBytesRepr(self.eventsData)
            names = [str(n) for (n, s) in self.sequences]
            return f'{type(self).__name__}({data}, names={names!r})'


class SSARSequence:
    """
    A sequence within a *SSAR* sequence archive file. These generally
    contain sound effects.
    """
    def __init__(self, firstEvent_firstEventOffset, bankID=0, volume=127,
                 channelPressure=64, polyphonicPressure=50, playerID=0,
                 *, parsed=True):

        if parsed:
            self._firstEvent = firstEvent_firstEventOffset
            self.firstEventOffset = 0
        else:
            self._firstEvent = None
            if firstEvent_firstEventOffset == -1:
                firstEvent_firstEventOffset = None
            self.firstEventOffset = firstEvent_firstEventOffset

        self.bankID = bankID
        self.volume = volume
        self.channelPressure = channelPressure
        self.polyphonicPressure = polyphonicPressure
        self.playerID = playerID
        self.parsed = parsed


    @property
    def firstEvent(self):
        if not self.parsed:
            raise ValueError('You must parse the SSAR with .parse() before you'
                             ' can access .firstEvent in sequences!')
        return self._firstEvent
    @firstEvent.setter
    def firstEvent(self, value):
        self._firstEvent = value


    @property
    def firstEventOffset(self):
        if self.parsed:
            raise ValueError('You cannot access .firstEventOffset after you'
                             ' have parsed the SSAR!')
        return self._firstEventOffset
    @firstEventOffset.setter
    def firstEventOffset(self, value):
        self._firstEventOffset = value


    def save(self):
        """
        Return this SSAR sequence's first event or first event offset,
        .bankID, .volume, .channelPressure, .polyphonicPressure, and
        .playerID as a 6-tuple. This matches the parameters of the
        default class constructor.
        """
        return (self.firstEvent if self.parsed else self.firstEventOffset,
                self.bankID,
                self.volume,
                self.channelPressure,
                self.polyphonicPressure,
                self.playerID)


    def __str__(self):
        up = ' unparsed' if not self.parsed else ''
        fields = []
        fields.append(f'bank={self.bankID}')
        fields.append(f'volume={self.volume}')
        fields.append(f'playerID={self.playerID}')
        return f'<ssar-sequence{up} {" ".join(fields)}>'


    def __repr__(self):
        fields = []
        fields.append(repr(self.firstEvent if self.parsed else self.firstEventOffset))
        fields.append(repr(self.bankID))
        fields.append(repr(self.volume))
        fields.append(repr(self.channelPressure))
        fields.append(repr(self.polyphonicPressure))
        fields.append(repr(self.playerID))
        if not self.parsed:
            fields.append('parsed=False')
        return f'{type(self).__name__}({", ".join(fields)})'



# Copyright 2019 RoadrunnerWMC
#
# This file is part of ndspy.
#
# ndspy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# ndspy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with ndspy.  If not, see <https://www.gnu.org/licenses/>.
"""
Support for SBNK instrument banks.
"""


import struct

from . import _common


# Constants for various instrument types
NO_INSTRUMENT_TYPE = 0
SINGLE_NOTE_PCM_INSTRUMENT_TYPE = 1
SINGLE_NOTE_PSG_SQUARE_WAVE_INSTRUMENT_TYPE = 2
SINGLE_NOTE_PSG_WHITE_NOISE_INSTRUMENT_TYPE = 3
RANGE_INSTRUMENT_TYPE = 16
REGIONAL_INSTRUMENT_TYPE = 17


class NoteDefinition:
    """
    A note definition within a SBNK instrument.
    """
    def __init__(self, waveID_dutyCycle=0, waveArchiveIDID=0, pitch=60,
            attack=127, decay=127, sustain=127, release=127, pan=64,
            unknown=1):
        self.waveID = waveID_dutyCycle
        self.waveArchiveIDID = waveArchiveIDID
        self.pitch = pitch
        self.attack = attack
        self.decay = decay
        self.sustain = sustain
        self.release = release
        self.pan = pan
        self.unknown = unknown


    @property
    def dutyCycle(self):
        return self.waveID
    @dutyCycle.setter
    def dutyCycle(self, value):
        self.waveID = value
    

    @classmethod
    def fromData(cls, data):
        """
        Create a note definition from raw file data that does not
        include the "unknown" value.
        """
        return cls(*struct.unpack_from('<HH6B', data), 1)


    @classmethod
    def fromDataWithUnknown(cls, data):
        """
        Create a note definition from raw file data that includes the
        "unknown" value.
        """
        values = struct.unpack_from('<3H6B', data)
        return cls(*values[1:], values[0])


    def save(self):
        """
        Generate data representing this note definition, without
        including the "unknown" value.
        """
        return struct.pack('<HH6B', self.waveID, self.waveArchiveIDID,
            self.pitch, self.attack, self.decay, self.sustain,
            self.release, self.pan)


    def saveWithUnknown(self):
        """
        Generate data representing this note definition, including the
        "unknown" value.
        """
        return struct.pack('<H', self.unknown) + self.save()


    def __str__(self, typeContext=SINGLE_NOTE_PCM_INSTRUMENT_TYPE):
        name = _common.noteName(self.pitch)

        if typeContext == SINGLE_NOTE_PSG_SQUARE_WAVE_INSTRUMENT_TYPE:
            pct = ['12.5%',
                   '25%',
                   '37.5%',
                   '50%',
                   '62.5%',
                   '75%',
                   '87.5%',
                   '0%'][self.dutyCycle & 7]
            extraInfo = f' ({pct} duty cycle)'
        elif typeContext == SINGLE_NOTE_PSG_WHITE_NOISE_INSTRUMENT_TYPE:
            extraInfo = ''
        else:
            extraInfo = f' (SWAR {self.waveArchiveIDID}, SWAV {self.waveID})'

        return f'<note def {name}{extraInfo}>'


    def __repr__(self):
        params = ', '.join(repr(s) for s in
            [self.waveID, self.waveArchiveIDID, self.pitch, self.attack,
            self.decay, self.sustain, self.release, self.pan,
            self.unknown])
        return f'{type(self).__name__}({params})'
        


class Instrument:
    """
    An instrument within a SBNK file. This is an abstract base class,
    and should be subclassed in order to be used.
    """

    bankOrderKey = 0
    dataMergeOptimizationID = 0


    def __init__(self, type):
        """
        Initialize the instrument.
        """
        self.type = type


    @classmethod
    def fromData(cls, type, data, startOffset):
        """
        Create an instrument from raw file data.

        This method must be implemented in subclasses; this
        abstract-base-class implementation simply raises
        NotImplementedError.
        """
        raise NotImplementedError('Override fromData() in Instrument '
                                  'subclasses!')


    def save(self):
        """
        Return the instrument's type value as a 1-tuple. Subclasses may
        return longer tuples with more data; currently, all subclasses
        add a :py:class:`bytes` instance.
        """
        return (self.type,)


class SingleNoteInstrument(Instrument):
    """
    An instrument that contains one note definition and nothing else.
    This class encompasses instrument type values 1 through 15.
    """
    def __init__(self, type, noteDefinition):
        super().__init__(type)
        self.noteDefinition = noteDefinition


    @classmethod
    def fromData(cls, type, data, startOffset):
        """
        Create a single-note instrument from raw file data.
        """

        noteData = data[startOffset : startOffset + 10]
        noteDefinition = NoteDefinition.fromData(noteData)

        return cls(type, noteDefinition), 10


    def save(self):
        """
        Generate file data representing this instrument, and then return
        the instrument's type value and that data as a pair.
        """
        return super().save() + (self.noteDefinition.save(),)


    def __str__(self):

        if self.type == SINGLE_NOTE_PCM_INSTRUMENT_TYPE:
            typeName = 'PCM'
        elif self.type == SINGLE_NOTE_PSG_SQUARE_WAVE_INSTRUMENT_TYPE:
            typeName = 'PSG square wave'
        elif self.type == SINGLE_NOTE_PSG_WHITE_NOISE_INSTRUMENT_TYPE:
            typeName = 'PSG white noise'
        else:
            typeName = 'unknown'

        try:
            nd = self.noteDefinition.__str__(self.type)
        except TypeError:
            # The object's __str__ method doesn't support the `type`
            # parameter, so it probably isn't NoteDefinition or a
            # subclass of it. In this case, just use str().
            nd = str(self.noteDefinition)

        return f'<single-note {typeName} instrument {nd}>'


    def __repr__(self):
        return (f'{type(self).__name__}('
                f'{self.type!r}, '
                f'{self.noteDefinition!r})')


class RangeInstrument(Instrument):
    """
    An instrument that contains one note definition for each pitch in a
    given range.
    """
    def __init__(self, firstPitch, noteDefinitions):
        super().__init__(RANGE_INSTRUMENT_TYPE)
        self.firstPitch = firstPitch
        self.noteDefinitions = noteDefinitions


    @classmethod
    def fromData(cls, _, data, startOffset):
        """
        Create a range instrument from raw file data.
        """
        firstPitch, lastPitch = \
            struct.unpack_from('<BB', data, startOffset)

        noteDefinitions = []
        off = startOffset + 2
        for i in range(lastPitch - firstPitch + 1):
            noteData = data[off : off+0xC]
            noteDefinitions.append(
                NoteDefinition.fromDataWithUnknown(noteData))
            off += 0xC

        return (cls(firstPitch, noteDefinitions),
                off - startOffset)


    def save(self):
        """
        Generate file data representing this instrument, and then return
        the instrument's type value and that data as a pair.
        """
        data = bytearray(2 + 0xC * len(self.noteDefinitions))
        struct.pack_into('<BB', data, 0,
            self.firstPitch, self.firstPitch + len(self.noteDefinitions) - 1)

        off = 2
        for n in self.noteDefinitions:
            data[off : off+0xC] = n.saveWithUnknown()
            off += 0xC

        return super().save() + (data,)


    def __str__(self):
        joinList = []
        for i, d in enumerate(self.noteDefinitions):
            joinList.append(
                f'{_common.noteName(i + self.firstPitch)}: {d}')
        defs = ', '.join(joinList)
        firstNote = _common.noteName(self.firstPitch)
        lastNote = _common.noteName(self.lastPitch)
        return f'<range instrument {firstNote}-{lastNote} {{{defs}}}>'


    def __repr__(self):
        defs = ', '.join(repr(d) for d in self.noteDefinitions)
        return (f'{type(self).__name__}('
                f'{self.firstPitch!r}, '
                f'{self.lastPitch!r}, '
                f'[{defs}])')


class RegionalInstrument(Instrument):
    """
    An instrument that partitions the range [0, 127] into sections, and
    contains one note definition for each.
    """

    class Region:
        """
        A region within a regional instrument.
        """
        def __init__(self, lastPitch, noteDefinition):
            self.lastPitch = lastPitch
            self.noteDefinition = noteDefinition


        def __str__(self):
            return (f'<region through {_common.noteName(self.lastPitch)} '
                    f'{self.noteDefinition}>')


        def __repr__(self):
            return (f'{type(self).__name__}('
                    f'{self.lastPitch!r}, '
                    f'{self.noteDefinition!r})')


    def __init__(self, regions):
        super().__init__(REGIONAL_INSTRUMENT_TYPE)
        self.regions = regions


    @classmethod
    def fromData(cls, _, data, startOffset):
        """
        Create a regional instrument from raw file data.
        """
        regionEnds = struct.unpack_from('8B', data, startOffset)

        regions = []
        off = startOffset + 8
        for e in regionEnds:
            if e == 0 and off != startOffset + 8: break

            noteData = data[off : off+0xC]
            note = NoteDefinition.fromDataWithUnknown(noteData)
            off += 0xC

            regions.append(cls.Region(e, note))

        return cls(regions), off - startOffset


    def save(self):
        """
        Generate file data representing this instrument, and then return
        the instrument's type value and that data as a pair.
        """
        if len(self.regions) > 8:
            raise ValueError("Can't save RegionalInstrument -- too many "
                             f'regions ({len(self.regions)})!')

        data = bytearray(8 + 0xC * len(self.regions))
        for i, region in enumerate(self.regions):
            data[i] = region.lastPitch
            data[8 + 0xC * i : 0x14 + 0xC * i] = \
                region.noteDefinition.saveWithUnknown()

        return super().save() + (data,)


    def __str__(self):
        fill = []
        last = 0
        for region in self.regions:
            def betterNoteName(num):
                if num == 0: return 'min'
                elif num == 127: return 'max'
                else: return _common.noteName(num)
            fill.append(f'{betterNoteName(last)}'
                        f'-{betterNoteName(region.lastPitch)}: '
                        f'{region.noteDefinition}')
            last = region.lastPitch + 1
        return f'<regional instrument {{{", ".join(fill)}}}>'


    def __repr__(self):
        return (f'{type(self).__name__}('
                f'[{", ".join(repr(r) for r in self.regions)}])')


def instrumentClass(type):
    """
    A convenience function that returns the Instrument subclass that
    should be used to load an instrument with the given type value.
    """
    if type == NO_INSTRUMENT_TYPE: # 0
        return None
    elif type < RANGE_INSTRUMENT_TYPE: # < 16
        return SingleNoteInstrument
    elif type == RANGE_INSTRUMENT_TYPE: # 16
        return RangeInstrument
    elif type == REGIONAL_INSTRUMENT_TYPE: # 17
        return RegionalInstrument
    else:
        raise ValueError(f'Instrument type {type} is invalid.')


def guessInstrumentType(data, startOffset, possibleTypes, bytesAvailable):
    """
    Try to guess the type of instrument stored in some binary data based
    on both the data and a set of possible types (ones that haven't been
    ruled out by the instrument's position in the surrounding data).
    This function is entirely based on heuristics, so it may return
    different answers for similar data, and it cannot always be
    accurate. Types 1, 2 and 3 are considered equivalent by this
    function, since they are very similar and all use the same Python
    class (SingleNoteInstrument). None will be returned if it's very
    unlikely that there is an instrument at that position.
    """
    possibleTypes = set(possibleTypes)

    # Types 2 and 3 (the PSG single-note instruments) should be
    # considered equivalent to type 1 (PCM single-note instrument)
    if SINGLE_NOTE_PSG_SQUARE_WAVE_INSTRUMENT_TYPE in possibleTypes:
        possibleTypes.remove(SINGLE_NOTE_PSG_SQUARE_WAVE_INSTRUMENT_TYPE)
        possibleTypes.add(SINGLE_NOTE_PCM_INSTRUMENT_TYPE)
    if SINGLE_NOTE_PSG_WHITE_NOISE_INSTRUMENT_TYPE in possibleTypes:
        possibleTypes.remove(SINGLE_NOTE_PSG_WHITE_NOISE_INSTRUMENT_TYPE)
        possibleTypes.add(SINGLE_NOTE_PCM_INSTRUMENT_TYPE)

    # By doing
    # if returnEarly(): return early()
    # you can quickly check if len(possibleTypes) < 2, and return the
    # appropriate type or None.
    def returnEarly():
        return len(possibleTypes) < 2
    def early():
        if possibleTypes:
            return list(possibleTypes)[0]
        else:
            return None


    # Return immediately if possible
    if returnEarly(): return early()


    # Start by ruling out the impossible...
    def ruleOut(type):
        if type in possibleTypes:
            possibleTypes.remove(type)
    if bytesAvailable < 10:
        ruleOut(NO_INSTRUMENT_TYPE)
    if bytesAvailable < 2 + 0xC:
        ruleOut(RANGE_INSTRUMENT_TYPE)
    if bytesAvailable < 8 + 0xC:
        ruleOut(REGIONAL_INSTRUMENT_TYPE)

    if returnEarly(): return early()


    # Then rule out the improbable...

    if SINGLE_NOTE_PCM_INSTRUMENT_TYPE in possibleTypes:
        # SWAV ID and SWAR ID will probably never be > 0x400... right?
        if data[startOffset + 1] > 4: ruleOut(SINGLE_NOTE_PCM_INSTRUMENT_TYPE)
        if data[startOffset + 3] > 4: ruleOut(SINGLE_NOTE_PCM_INSTRUMENT_TYPE)
        # And pitch will probably never be 0, right?
        if data[startOffset + 4] == 0: ruleOut(SINGLE_NOTE_PCM_INSTRUMENT_TYPE)
        # For that matter, if pitch is 0x3C (middle C), we can probably
        # just assume straight away that this is a single-note inst
        if data[startOffset + 4] == 0x3C:
            return SINGLE_NOTE_PCM_INSTRUMENT_TYPE
    if returnEarly(): return early()

    if RANGE_INSTRUMENT_TYPE in possibleTypes:
        firstPitch, lastPitch = data[startOffset : startOffset+2]
        if firstPitch > lastPitch:
            ruleOut(RANGE_INSTRUMENT_TYPE)
        else:
            expectedLen = 2 + 0xC * (lastPitch - firstPitch + 1)
            if expectedLen > bytesAvailable:
                ruleOut(RANGE_INSTRUMENT_TYPE)
    if returnEarly(): return early()

    if REGIONAL_INSTRUMENT_TYPE in possibleTypes:
        regionEnds = data[startOffset : startOffset+8]

        # Check that all ends are strictly increasing, followed by
        # all zeroes
        # (not every regional instrument ends with 7F, surprisingly)
        previous = -1
        for end in regionEnds:
            if previous != 0 and end == 0:
                # first zero byte
                previous = 0
            elif previous == 0 and end != 0:
                # nonzero byte after a zero byte? bad bad bad
                ruleOut(REGIONAL_INSTRUMENT_TYPE)
                break
            elif previous != 0:
                if end <= previous:
                    ruleOut(REGIONAL_INSTRUMENT_TYPE)
                    break
                previous = end
        else:
            # Finally, check that we have sufficient length
            regionCount = 0
            for e in regionEnds:
                if not e: break
                regionCount += 1
            expectedLen = 8 + 0xC * regionCount
            if expectedLen > bytesAvailable:
                ruleOut(REGIONAL_INSTRUMENT_TYPE)
    if returnEarly(): return early()


    # At this point, just pick one of the remaining options at random.
    return early()


class SBNK:
    """
    A SBNK instrument bank file. This defines a set of instruments
    that sequences and sequence archives can use.
    """
    # When saving SDAT, two otherwise identical SBNKs will share data
    # only if their dataMergeOptimizationIDs are the same.
    # You can pretty safely ignore this.
    dataMergeOptimizationID = 0

    def __init__(self, file=None, unk02=0, waveArchiveIDs=None):
        self.unk02 = unk02

        if waveArchiveIDs is None:
            self.waveArchiveIDs = []
        else:
            self.waveArchiveIDs = list(waveArchiveIDs)

            # Remove trailing `None`s if they exist
            while self.waveArchiveIDs and self.waveArchiveIDs[-1] is None:
                self.waveArchiveIDs = self.waveArchiveIDs[:-1]

        self.instruments = []
        self.inaccessibleInstruments = {}

        if file is not None:
            if not file.startswith(b'SBNK'):
                raise ValueError("Wrong magic (should be b'SBNK', instead "
                                 f'found {file[:4]})')

            self._initFromData(file)


    @classmethod
    def fromInstruments(cls, instruments, unk02=0, waveArchiveIDs=None):
        """
        Create a SBNK from a list of instruments.
        """
        obj = cls(unk02=unk02, waveArchiveIDs=waveArchiveIDs)
        obj.instruments = instruments
        return obj


    @classmethod
    def fromFile(cls, filePath, *args, **kwargs):
        """
        Load an SBNK from a filesystem file.
        """
        with open(filePath, 'rb') as f:
            return cls(f.read(), *args, **kwargs)
    

    def _initFromData(self, data):
        """
        Initialize the SBNK from file data.
        """

        magic, bom, version, filesize, headersize, numblocks = \
            _common.NDS_STD_FILE_HEADER.unpack_from(data, 0)
        assert magic == b'SBNK'
        if version != 0x100:
            raise ValueError(f'Unsupported SBNK version: {version}')

        dataMagic, dataSize, instrumentCount = \
            struct.unpack_from('<4sI32xI', data, 0x10)
        assert dataMagic == b'DATA'

        unconsumedBytes = set(range(0x3C + instrumentCount * 4, filesize))

        idsToOffsets = {}

        def makeInstrumentAt(type, offset):
            instrumentCls = instrumentClass(type)
            if instrumentCls is None:
                instrument = None
                consumedBytesHere = 0
            else:
                instrument, consumedBytesHere = instrumentCls.fromData(type,
                                                                       data,
                                                                       offset)
                instrument.bankOrderKey = offset
                instrument.dataMergeOptimizationID = offset

            for j in range(consumedBytesHere):
                if offset + j in unconsumedBytes:
                    unconsumedBytes.remove(offset + j)

            return instrument

        dataArrayPos = 0x3C
        for id in range(instrumentCount):
            (instrumentType, instrumentOffset) = \
                struct.unpack_from('<BHx', data, dataArrayPos)
            dataArrayPos += 4

            instrument = makeInstrumentAt(instrumentType, instrumentOffset)
            self.instruments.append(instrument)

            if instrument is not None:
                idsToOffsets[id] = instrumentOffset

        # Why do we have to do this D:
        while unconsumedBytes:

            # Account for possible 2 bytes of padding at the end
            if (filesize - 1 in unconsumedBytes
                and filesize - 2 in unconsumedBytes
                and filesize - 3 not in unconsumedBytes):
                unconsumedBytes.remove(filesize - 1)
                unconsumedBytes.remove(filesize - 2)
            if not unconsumedBytes: break

            thisOffset = min(unconsumedBytes)

            # Find the previous and following instrument IDs
            prevID = -1
            prevOffset = -1
            nextID = 9E9
            nextOffset = 9E9
            for id, offset in idsToOffsets.items():
                if offset < thisOffset and offset > prevOffset:
                    prevID = id
                    prevOffset = offset
                if offset > thisOffset and offset < nextOffset:
                    nextID = id
                    nextOffset = offset
            if prevID == -1: prevID = None
            if nextID == 9E9: nextID = None

            # Find possible types for this instrument based on that
            possibleTypes = {SINGLE_NOTE_PCM_INSTRUMENT_TYPE,
                             RANGE_INSTRUMENT_TYPE,
                             REGIONAL_INSTRUMENT_TYPE}
            prevInst = None if prevID is None else self.instruments[prevID]
            nextInst = None if nextID is None else self.instruments[nextID]
            if prevInst is not None:
                if prevInst.type >= RANGE_INSTRUMENT_TYPE:
                    possibleTypes.remove(1)
                if prevInst.type == REGIONAL_INSTRUMENT_TYPE:
                    possibleTypes.remove(RANGE_INSTRUMENT_TYPE)
            if nextInst is not None:
                if (nextInst.type <= RANGE_INSTRUMENT_TYPE
                        and REGIONAL_INSTRUMENT_TYPE in possibleTypes):
                    possibleTypes.remove(REGIONAL_INSTRUMENT_TYPE)
                if (nextInst.type < RANGE_INSTRUMENT_TYPE
                        and RANGE_INSTRUMENT_TYPE in possibleTypes):
                    possibleTypes.remove(RANGE_INSTRUMENT_TYPE)

            # Count the number of contiguous bytes we've got
            tempOffset = thisOffset + 1
            while tempOffset in unconsumedBytes:
                tempOffset += 1
            bytesAvailable = tempOffset - thisOffset

            # Use advanced heuristics™ to determine this instrument's
            # type
            instrumentType = guessInstrumentType(data,
                                                 thisOffset,
                                                 possibleTypes,
                                                 bytesAvailable)

            if instrumentType is None:
                instrument = None
            else:
                try:
                    instrument = makeInstrumentAt(instrumentType, thisOffset)
                except Exception:
                    instrument = None

            if instrument is None:
                # Couldn't make an instrument there, so, assume that
                # it's garbage data that we have to skip (to avoid an
                # infinite loop)
                unconsumedBytes.remove(thisOffset)
                if thisOffset + 1 in unconsumedBytes:
                    unconsumedBytes.remove(thisOffset + 1)

            else:
                if prevID not in self.inaccessibleInstruments:
                    self.inaccessibleInstruments[prevID] = []
                self.inaccessibleInstruments[prevID].append(instrument)


    def save(self):
        """
        Generate file data representing this SBNK, and then return that
        data, .unk02, and .waveArchiveIDs as a triple.
        """
        data = bytearray(0x3C + 4 * len(self.instruments))
        instrumentsData = bytearray()

        # Instrument data (that the offset points to) is stored in order of:
        # - Single-note instruments
        # - Range instruments
        # - Regional instruments
        # Yet the instruments offsets table is in the order that forms
        # the instruments IDs exposed to other files. So we save in two
        # steps.
        # Also, the order within each of those three categories can be
        # arbitrary. So instruments have a bankOrderKey that can be used
        # to sort within each category.

        # First, put together the instruments data itself
        indexToOffset = [None for _ in self.instruments]
        dataReuseCache = {}
        def addInstrument(instrument):
            instrumentType, instrumentData = instrument.save()
            instrumentData = bytes(instrumentData)
            mergeID = instrument.dataMergeOptimizationID

            if (instrumentData, mergeID) not in dataReuseCache:
                dataReuseCache[(instrumentData, mergeID)] = \
                    len(instrumentsData)
                instrumentsData.extend(instrumentData)

            return dataReuseCache[(instrumentData, mergeID)]

        test1 = lambda id: id < RANGE_INSTRUMENT_TYPE
        test2 = lambda id: id == RANGE_INSTRUMENT_TYPE
        test3 = lambda id: id == REGIONAL_INSTRUMENT_TYPE
        if None in self.inaccessibleInstruments:
            for inacc in self.inaccessibleInstruments[None]:
                addInstrument(inacc)
        for test in (test1, test2, test3):
            inThisGroup = []
            for i, instrument in enumerate(self.instruments):
                if instrument is None: continue
                if test(instrument.type):
                    inThisGroup.append((i, instrument))

            inThisGroup.sort(key=lambda elem: elem[1].bankOrderKey)

            for i, instrument in inThisGroup:
                indexToOffset[i] = addInstrument(instrument)

                if i in self.inaccessibleInstruments:
                    for inacc in self.inaccessibleInstruments[i]:
                        addInstrument(inacc)

        # Add any inaccessible instruments with IDs that are too high
        idSet = set(self.inaccessibleInstruments)
        if None in idSet: idSet.remove(None)
        for inaccID in sorted(idSet):
            if inaccID < len(self.instruments): continue
            for inacc in self.inaccessibleInstruments[inaccID]:
                addInstrument(inacc)

        # And now construct the table using that.
        for i, instrument in enumerate(self.instruments):
            instrumentType = instrument.type if instrument is not None else 0

            offs = indexToOffset[i]
            if offs is None: offs = 0
            else: offs += len(data)

            struct.pack_into('<BH', data, 0x3C + 4 * i,
                             instrumentType, offs)

        data.extend(instrumentsData)

        while len(data) % 4:
            data.append(0)

        _common.NDS_STD_FILE_HEADER.pack_into(data, 0,
            b'SBNK', 0xFEFF, 0x100, len(data), 0x10, 1)
        struct.pack_into('<4sI32xI', data, 0x10,
            b'DATA', len(data) - 0x10, len(self.instruments))

        return (data, self.unk02, self.waveArchiveIDs)


    def saveToFile(self, filePath):
        """
        Generate file data representing this SBNK, and save it to a
        filesystem file.
        """
        d = self.save()[0]
        with open(filePath, 'wb') as f:
            f.write(d)


    def __str__(self):
        linesList = [f'<sbnk waveArchiveIDs={self.waveArchiveIDs}']

        maxNumLen = len(str(len(self.instruments) - 1))

        for i, inst in enumerate(self.instruments):
            linesList.append(f'    [{i:{maxNumLen}}] {inst}')

        linesList.append('>')
        return '\n'.join(linesList)


    def __repr__(self):
        return (f'{type(self).__name__}.fromInstruments('
                f'{self.instruments!r}, '
                f'waveArchiveIDs={self.waveArchiveIDs!r})')



# Copyright 2019 RoadrunnerWMC
#
# This file is part of ndspy.
#
# ndspy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# ndspy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with ndspy.  If not, see <https://www.gnu.org/licenses/>.


import struct

from . import _common
from . import texture


class NSBMD:
    """
    A 3D model.
    """
    def __init__(self, data=None):
        if data is not None:
            self._initFromData(data)


    def _initFromData(self, data):
        """
        Load the NSBMD from file data.
        """
        if not data.startswith(b'BMD0'):
            raise ValueError('Invalid NSBMD: incorrect magic')

        magic, bom, version, filesize, headersize, numblocks = \
            _common.NDS_STD_FILE_HEADER.unpack_from(data, 0)
        if version != 2:
            raise ValueError(f'Unsupported NSBMD version: {version}')

        blockOffsets = struct.unpack_from(f'<{numblocks}I', data, 0x10)
        for off in blockOffsets:
            blockMagic, blockLen = struct.unpack_from('<4sI', data, off)
            blockData = data[off : off+blockLen]
            print('BLOCK OFF ' + blockMagic.decode('ascii') + ' ' + hex(off))

            if blockMagic == b'MDL0':
                self._parseMDL0(blockData)
            elif blockMagic == b'TEX0':
                self._parseTEX0(blockData)
            else:
                raise ValueError(f'Unknown NSBMD block: {blockMagic}')


    def _parseMDL0(self, data):
        """
        Parse the MDL0 block.
        """
        assert data.startswith(b'MDL0')
        
        modelsCount, info3dSize = struct.unpack_from('<xBH', data, 0x8)

        off = 0x0C
        unkBlockHeaderLen, unkBlockLen, unkBlockUnk04 = struct.unpack_from('<HHI', data, off)
        assert unkBlockHeaderLen == 8
        assert unkBlockUnk04 == 0x17F
        unkBlockEntries = []
        for i in range(modelsCount):
            unk00, unk02 = struct.unpack_from('<HH', data, off + 0x8 + 4 * i)
            unkBlockEntries.append((unk00, unk02))

        off += 8 + 4 * modelsCount # unkBlockLen is a lie?
        infoSectionLen, infoLen = struct.unpack_from('<HH', data, off)
        modelOffsets = []
        for i in range(modelsCount):
            modelOff, = struct.unpack_from('<I', data, off + 0x4 + 4 * i)
            modelOffsets.append(modelOff)

        off += 4 + 4 * modelsCount
        modelNames = []
        for i in range(modelsCount):
            name = data[off : off+16].rstrip(b'\0').decode('latin-1')
            modelNames.append(name)
            off += 16

        self.models = []
        for (unk00, unk02), modelOff, modelName in zip(unkBlockEntries, modelOffsets, modelNames):
            (modelSize, bonesOffset, materialsOffset,
                    polygonsBeginOffset, polygonsEndOffset,
                    unk14, unk15, unk16, objectsCount, materialsCount,
                    polygonsCount, unk1A, unk1B, unk1C, unk1D, unk1E, unk1F,
                    unk20, scaleMode, unk22, unk23, verticesCount,
                    surfacesCount, trianglesCount, quadsCount,
                    boundingBoxX, boundingBoxY, boundingBoxZ,
                    boundingBoxWidth, boundingBoxHeight, boundingBoxDepth,
                    runtimeUse) = \
                struct.unpack_from('<5I16B4H6h8s', data, modelOff)
            print('RUNTIME USE @ ' + hex(modelOff + 0x38))
            print(runtimeUse)

            # Bones
            print('BONES @ ' + hex(modelOff + bonesOffset))

            # Materials
            print('MATERIALS @ ' + hex(modelOff + materialsOffset))
            off = modelOff + materialsOffset + 4 # TODO: WHY DO WE HAVE TO ADD 4
            materialsCount, materialsHeaderSize = struct.unpack_from('<xBH', data, off)
            off += 4

            unkMatBlockHeaderLen, unkMatBlockLen, unkMatBlockUnk04 = \
                struct.unpack_from('<HHI', data, off)
            off += 8
            assert unkMatBlockHeaderLen == 8
            assert unkMatBlockUnk04 == 0x17F
            unkMatBlockEntries = []
            for i in range(materialsCount):
                unk00, unk02 = struct.unpack_from('<HH', data, off)
                unkMatBlockEntries.append((unk00, unk02))
                off += 4

            matOffsetsHeaderLen, matOffsetsLen = \
                struct.unpack_from('<HH', data, off)
            off += 4
            assert matOffsetsHeaderLen == 4
            matOffsets = []
            for i in range(materialsCount):
                matOffsets.append(struct.unpack_from('<I', data, off)[0])
                off += 4

            matNames = []
            for i in range(materialsCount):
                name = data[off : off+16].rstrip(b'\0').decode('latin-1')
                matNames.append(name)
                off += 16

            texturesCount, texturesHeaderSize = struct.unpack_from('<xBH', data, off)
            off += 4

            unkTexBlockHeaderLen, unkTexBlockLen, unkTexBlockUnk04 = \
                struct.unpack_from('<HHI', data, off)
            off += 8
            assert unkTexBlockHeaderLen == 8
            assert unkTexBlockUnk04 == 0x17F
            unkTexBlockEntries = []
            for i in range(texturesCount):
                unk00, unk02 = struct.unpack_from('<HH', data, off)
                unkTexBlockEntries.append((unk00, unk02))
                off += 4

            texDatasHeaderLen, texDatasLen = \
                struct.unpack_from('<HH', data, off)
            off += 4
            assert texDatasHeaderLen == 4
            texDatas = []
            for i in range(texturesCount):
                texDatas.append(struct.unpack_from('<HBx', data, off))
                off += 4

            texNames = []
            for i in range(texturesCount):
                name = data[off : off+16].rstrip(b'\0').decode('latin-1')
                texNames.append(name)
                off += 16

            palettesCount, palettesHeaderSize = struct.unpack_from('<xBH', data, off)
            off += 4

            unkPalBlockHeaderLen, unkPalBlockLen, unkPalBlockUnk04 = \
                struct.unpack_from('<HHI', data, off)
            off += 8
            assert unkPalBlockHeaderLen == 8
            assert unkPalBlockUnk04 == 0x17F
            unkPalBlockEntries = []
            for i in range(palettesCount):
                unk00, unk02 = struct.unpack_from('<HH', data, off)
                unkPalBlockEntries.append((unk00, unk02))
                off += 4

            palDatasHeaderLen, palDatasLen = \
                struct.unpack_from('<HH', data, off)
            off += 4
            assert palDatasHeaderLen == 4
            palDatas = []
            for i in range(palettesCount):
                palDatas.append(struct.unpack_from('<HBx', data, off))
                off += 4

            palNames = []
            for i in range(palettesCount):
                name = data[off : off+16].rstrip(b'\0').decode('latin-1')
                palNames.append(name)
                off += 16

            print('Palettes:', palNames, palDatas, hex(off + 0x18))

            for (unk00, unk02), matOffset, matName in zip(unkMatBlockEntries, matOffsets, matNames):
                off = modelOff + materialsOffset + matOffset + 4
                print(hex(off + 0x18))


            # Polygons
            print('POLYGONS @ ' + hex(modelOff + polygonsBeginOffset))
            off = modelOff + polygonsBeginOffset
            polygonsCount, polygonsHeaderSize = struct.unpack_from('<xBH', data, off)
            off += 4

            unkPolyBlockHeaderLen, unkPolyBlockLen, unkPolyBlockUnk04 = \
                struct.unpack_from('<HHI', data, off)
            off += 8
            assert unkPolyBlockHeaderLen == 8
            assert unkPolyBlockUnk04 == 0x17F
            unkPolyBlockEntries = []
            for i in range(polygonsCount):
                unk00, unk02 = struct.unpack_from('<HH', data, off)
                unkPolyBlockEntries.append((unk00, unk02))
                off += 4

            polyOffsetsHeaderLen, polyOffsetsLen = \
                struct.unpack_from('<HH', data, off)
            off += 4
            assert polyOffsetsHeaderLen == 4
            polyOffsets = []
            for i in range(polygonsCount):
                polyOffsets.append(struct.unpack_from('<I', data, off)[0])
                off += 4

            polyNames = []
            for i in range(polygonsCount):
                name = data[off : off+16].rstrip(b'\0').decode('latin-1')
                polyNames.append(name)
                off += 16

            for (unk00, unk02), polyOffset, polyName in zip(unkPolyBlockEntries, polyOffsets, polyNames):
                polyDefOffset = modelOff + polygonsBeginOffset + polyOffset
                unkDef00, unkDef04, displayListOff, displayListLen = \
                    struct.unpack_from('<4I', data, polyDefOffset)

                displayListData = data[polyDefOffset + displayListOff : polyDefOffset + displayListOff + displayListLen]
                displayList = self._parseDisplayList(displayListData)
                # for x in displayList: print(x)
                # print('--------')

            model = None
            self.models.append((modelName, model))

        #print(self.models)

        self._MDL0 = data


    @staticmethod
    def _parseDisplayList(data):
        """
        Given display list data, convert it into a list of geometry
        commands.
        """
        commands = []
        i = 0
        while i < len(data):
            commandIDs = data[i : i+4]
            i += 4
            for id in commandIDs:
                cmdType = _COMMANDS_BY_ID.get(id)
                if cmdType is None:
                    print(f'WARNING: Unknown command ID {hex(id)}!')
                    cmdType = GeometryCommand
                    paramCount = 0
                else:
                    paramCount = cmdType.PARAM_COUNT
                params = []
                for j in range(paramCount):
                    params.append(data[i : i+4])
                    i += 4
                cmd = cmdType.fromData(params)
                cmd.type = cmdType # in case this command has an unknown ID
                commands.append(cmd)
        return commands


    def _parseTEX0(self, data):
        """
        Load the TEX0 block.
        """
        assert data.startswith(b'TEX0')

        parsed = texture._readTEX0(data)
        self.texUnk08 = parsed['unk08']
        self.texUnk10 = parsed['unk10']
        self.texUnk18 = parsed['unk18']
        self.texUnk20 = parsed['unk20']
        self.texUnk2C = parsed['unk2C']
        self.texUnk32 = parsed['unk32']
        self.textures = parsed['textures']
        self.palettes = parsed['palettes']


    def _saveMDL0(self):
        """
        Save the MDL0 block.
        """
        return self._MDL0


    def _saveTEX0(self):
        """
        Save the TEX0 block.
        """
        texInfo = {}
        texInfo['unk08'] = self.texUnk08
        texInfo['unk10'] = self.texUnk10
        texInfo['unk18'] = self.texUnk18
        texInfo['unk20'] = self.texUnk20
        texInfo['unk2C'] = self.texUnk2C
        texInfo['unk32'] = self.texUnk32
        texInfo['textures'] = self.textures
        texInfo['palettes'] = self.palettes
        return texture._saveTEX0(texInfo)


    def save(self):
        """
        Save the model back to a file.
        """

        data = bytearray(0x10)

        # Gather data blocks
        dataBlocks = []
        dataBlocks.append(self._saveMDL0())
        dataBlocks.append(self._saveTEX0())

        # Add space for the block offsets
        data.extend(b'\0' * (len(dataBlocks) * 4))

        # Add the block datas and insert the offsets into the table
        for i, block in enumerate(dataBlocks):
            struct.pack_into('<I', data, 0x10 + 4 * i, len(data))
            data.extend(block)

        # Add the NDS standard file header
        _common.NDS_STD_FILE_HEADER.pack_into(data, 0,
            b'BMD0', 0xFEFF, 2, len(data), 0x10, 2)

        # And return.
        return bytes(data)


    def exportNSBTX(self):
        """
        Export a NSBTX object containing the data that would be stored
        in this model's TEX0 block.
        """
        return texture.NSBTX.fromTexturesAndPalettes(self.texUnk08,
                                                     self.texUnk10,
                                                     self.texUnk18,
                                                     self.texUnk20,
                                                     self.texUnk2C,
                                                     self.texUnk32,
                                                     self.textures,
                                                     self.palettes)


    def importNSBTX(self, nsbtx):
        """
        Import a NSBTX object, using it to overwrite this model's TEX0
        block.
        """
        self.texUnk08 = nsbtx.unk08
        self.texUnk10 = nsbtx.unk10
        self.texUnk18 = nsbtx.unk18
        self.texUnk20 = nsbtx.unk20
        self.texUnk2C = nsbtx.unk2C
        self.texUnk32 = nsbtx.unk32
        self.textures = nsbtx.textures
        self.palettes = nsbtx.palettes



class GeometryCommand():
    type = None
    PARAM_COUNT = 0

    @classmethod
    def fromData(cls, params):
        return cls()


class GeometryCommand_Nop(GeometryCommand):
    type = 0


class GeometryCommand_SetMatrixMode(GeometryCommand):
    """
    Also known as MTX_MODE.
    """
    type = 0x10
    PARAM_COUNT = 1


class GeometryCommand_PushMatrix(GeometryCommand):
    """
    Also known as MTX_PUSH.
    """
    type = 0x11
    PARAM_COUNT = 0


class GeometryCommand_PopMatrix(GeometryCommand):
    """
    Also known as MTX_POP.
    """
    type = 0x12
    PARAM_COUNT = 1


class GeometryCommand_StoreMatrix(GeometryCommand):
    """
    Also known as MTX_STORE.
    """
    type = 0x13
    PARAM_COUNT = 1


class GeometryCommand_RestoreMatrix(GeometryCommand):
    """
    Also known as MTX_RESTORE.
    """
    type = 0x14
    PARAM_COUNT = 1


class GeometryCommand_LoadIdentityMatrix(GeometryCommand):
    """
    Also known as MTX_IDENTITY.
    """
    type = 0x15
    PARAM_COUNT = 0


class GeometryCommand_Load4x4Matrix(GeometryCommand):
    """
    Also known as MTX_LOAD_4x4.
    """
    type = 0x16
    PARAM_COUNT = 16


class GeometryCommand_Load4x3Matrix(GeometryCommand):
    """
    Also known as MTX_LOAD_4x3.
    """
    type = 0x17
    PARAM_COUNT = 12


class GeometryCommand_MultiplyBy4x4Matrix(GeometryCommand):
    """
    Also known as MTX_MULT_4x4.
    """
    type = 0x18
    PARAM_COUNT = 16


class GeometryCommand_MultiplyBy4x3Matrix(GeometryCommand):
    """
    Also known as MTX_MULT_4x3.
    """
    type = 0x19
    PARAM_COUNT = 12


class GeometryCommand_MultiplyBy3x3Matrix(GeometryCommand):
    """
    Also known as MTX_MULT_3x3.
    """
    type = 0x1A
    PARAM_COUNT = 9


class GeometryCommand_MultiplyByScaleMatrix(GeometryCommand):
    """
    Also known as MTX_SCALE.
    """
    type = 0x1B
    PARAM_COUNT = 3


class GeometryCommand_MultiplyByTranslationMatrix(GeometryCommand):
    """
    Also known as MTX_TRANS.
    """
    type = 0x1C
    PARAM_COUNT = 3


class GeometryCommand_SetVertexColor(GeometryCommand):
    """
    Also known as COLOR.
    """
    type = 0x20
    PARAM_COUNT = 1


class GeometryCommand_SetNormalVector(GeometryCommand):
    """
    Also known as NORMAL.
    """
    type = 0x21
    PARAM_COUNT = 1


class GeometryCommand_SetTextureCoordinates(GeometryCommand):
    """
    Also known as TEXCOORD.
    """
    type = 0x22
    PARAM_COUNT = 1


class GeometryCommand_SetVertexXYZCoordinates_16BitsEach(GeometryCommand):
    """
    Also known as VTX_16.
    """
    type = 0x23
    PARAM_COUNT = 2


class GeometryCommand_SetVertexXYZCoordinates_10BitsEach(GeometryCommand):
    """
    Also known as VTX_10.
    """
    type = 0x24
    PARAM_COUNT = 1


class GeometryCommand_SetVertexXYCoordinates(GeometryCommand):
    """
    Also known as VTX_XY.
    """
    type = 0x25
    PARAM_COUNT = 1


class GeometryCommand_SetVertexXZCoordinates(GeometryCommand):
    """
    Also known as VTX_XZ.
    """
    type = 0x26
    PARAM_COUNT = 1


class GeometryCommand_SetVertexYZCoordinates(GeometryCommand):
    """
    Also known as VTX_YZ.
    """
    type = 0x27
    PARAM_COUNT = 1


class GeometryCommand_SetRelativeVertexCoordinates(GeometryCommand):
    """
    Also known as VTX_DIFF.
    """
    type = 0x28
    PARAM_COUNT = 1


class GeometryCommand_SetPolygonAttributes(GeometryCommand):
    """
    Also known as POLYGON_ATTR.
    """
    type = 0x29
    PARAM_COUNT = 1


class GeometryCommand_SetTextureParameters(GeometryCommand):
    """
    Also known as TEXIMAGE_PARAM.
    """
    type = 0x2A
    PARAM_COUNT = 1


class GeometryCommand_SetTexturePaletteBaseAddress(GeometryCommand):
    """
    Also known as PLTT_BASE.
    """
    type = 0x2B
    PARAM_COUNT = 1


class GeometryCommand_SetDiffuseAndAmbientReflection(GeometryCommand):
    """
    Also known as DIF_AMB.
    """
    type = 0x30
    PARAM_COUNT = 1


class GeometryCommand_SetSpecularReflectionAndEmission(GeometryCommand):
    """
    Also known as SPE_EMI.
    """
    type = 0x31
    PARAM_COUNT = 1


class GeometryCommand_SetLightDirectionalVector(GeometryCommand):
    """
    Also known as LIGHT_VECTOR.
    """
    type = 0x32
    PARAM_COUNT = 1


class GeometryCommand_SetLightColor(GeometryCommand):
    """
    Also known as LIGHT_COLOR.
    """
    type = 0x33
    PARAM_COUNT = 1


class GeometryCommand_SetSpecularReflectionShininessTable(GeometryCommand):
    """
    Also known as SHININESS.
    """
    type = 0x34
    PARAM_COUNT = 32


class GeometryCommand_StartOfVertexList(GeometryCommand):
    """
    Also known as BEGIN_VTXS.
    """
    type = 0x40
    PARAM_COUNT = 1


class GeometryCommand_EndOfVertexList(GeometryCommand):
    """
    Also known as END_VTXS.
    This seems to have no effect on hardware.
    """
    type = 0x41
    PARAM_COUNT = 0


class GeometryCommand_SwapRenderingEngineBuffers(GeometryCommand):
    """
    Also known as SWAP_BUFFERS.
    """
    type = 0x50
    PARAM_COUNT = 1


class GeometryCommand_SetViewport(GeometryCommand):
    """
    Also known as VIEWPORT.
    """
    type = 0x60
    PARAM_COUNT = 1


class GeometryCommand_TestIfCuboidSitsInsideViewVolume(GeometryCommand):
    """
    Also known as BOX_TEST.
    """
    type = 0x70
    PARAM_COUNT = 3


class GeometryCommand_SetPositionCoordinatesForTest(GeometryCommand):
    """
    Also known as POS_TEST.
    """
    type = 0x71
    PARAM_COUNT = 2


class GeometryCommand_SetDirectionalVectorForTest(GeometryCommand):
    """
    Also known as VEC_TEST.
    """
    type = 0x72
    PARAM_COUNT = 3


_COMMANDS_BY_ID = {
    0x00: GeometryCommand_Nop,
    0x10: GeometryCommand_SetMatrixMode,
    0x11: GeometryCommand_PushMatrix,
    0x12: GeometryCommand_PopMatrix,
    0x13: GeometryCommand_StoreMatrix,
    0x14: GeometryCommand_RestoreMatrix,
    0x15: GeometryCommand_LoadIdentityMatrix,
    0x16: GeometryCommand_Load4x4Matrix,
    0x17: GeometryCommand_Load4x3Matrix,
    0x18: GeometryCommand_MultiplyBy4x4Matrix,
    0x19: GeometryCommand_MultiplyBy4x3Matrix,
    0x1A: GeometryCommand_MultiplyBy3x3Matrix,
    0x1B: GeometryCommand_MultiplyByScaleMatrix,
    0x1C: GeometryCommand_MultiplyByTranslationMatrix,
    0x20: GeometryCommand_SetVertexColor,
    0x21: GeometryCommand_SetNormalVector,
    0x22: GeometryCommand_SetTextureCoordinates,
    0x23: GeometryCommand_SetVertexXYZCoordinates_16BitsEach,
    0x24: GeometryCommand_SetVertexXYZCoordinates_10BitsEach,
    0x25: GeometryCommand_SetVertexXYCoordinates,
    0x26: GeometryCommand_SetVertexXZCoordinates,
    0x27: GeometryCommand_SetVertexYZCoordinates,
    0x28: GeometryCommand_SetRelativeVertexCoordinates,
    0x29: GeometryCommand_SetPolygonAttributes,
    0x2A: GeometryCommand_SetTextureParameters,
    0x2B: GeometryCommand_SetTexturePaletteBaseAddress,
    0x30: GeometryCommand_SetDiffuseAndAmbientReflection,
    0x31: GeometryCommand_SetSpecularReflectionAndEmission,
    0x32: GeometryCommand_SetLightDirectionalVector,
    0x33: GeometryCommand_SetLightColor,
    0x34: GeometryCommand_SetSpecularReflectionShininessTable,
    0x40: GeometryCommand_StartOfVertexList,
    0x41: GeometryCommand_EndOfVertexList,
    0x50: GeometryCommand_SwapRenderingEngineBuffers,
    0x60: GeometryCommand_SetViewport,
    0x70: GeometryCommand_TestIfCuboidSitsInsideViewVolume,
    0x71: GeometryCommand_SetPositionCoordinatesForTest,
    0x72: GeometryCommand_SetDirectionalVectorForTest,
}



# Copyright 2019 RoadrunnerWMC
#
# This file is part of ndspy.
#
# ndspy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# ndspy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with ndspy.  If not, see <https://www.gnu.org/licenses/>.
"""
Support for BMG files.
"""


import struct

from . import _common


class BMG:
    """
    A class representing a BMG file.
    """

    def __init__(self, data=None, *, id=0):

        self.messages = []
        self.instructions = []
        self.labels = []
        self.scripts = []

        self.id = id

        self.unk10 = 2
        self.unk14 = 0
        self.unk18 = 0
        self.unk1C = 0

        if data is not None:
            self._initFromData(data)


    def _initFromData(self, data):
        if data[:8] != b'MESGbmg1':
            raise ValueError('Not a BMG file.')

        magic, dataLen, sectionCount, self.unk10, unk14, unk18, unk1C = \
            struct.unpack_from('<8s6I', data, 0)

        INF1 = []
        def parseINF1(offset, length):
            count, entryLength, self.id = struct.unpack_from('<HHI', data, offset + 8)
            assert entryLength == 8

            for i in range(count):
                entryOff, entryAttribs = struct.unpack_from('<II', data, offset + 16 + i * 8)
                INF1.append((entryOff, entryAttribs))

        DAT1 = b''
        def parseDAT1(offset, length):
            nonlocal DAT1
            DAT1 = data[offset + 8 : offset + length]

        self.instructions = []
        self.labels = []
        def parseFLW1(offset, length):
            instructionsCount, labelsCount, unk0C = \
                struct.unpack_from('<HHI', data, offset + 8)
            # unk0C is always 0, as far as I can tell

            instructionsTableOffset = offset + 16
            for i in range(instructionsCount):
                instOff = instructionsTableOffset + i * 8
                cmd = data[instOff : instOff + 8]
                if cmd != b'\0\0\0\0\0\0\0\0':
                    self.instructions.append(cmd)

            indicesTableOffset = instructionsTableOffset + instructionsCount * 8
            bmgIDsTableOffset = indicesTableOffset + labelsCount * 2
            for i in range(labelsCount):
                index, = struct.unpack_from('<h', data, indicesTableOffset + i * 2)
                bmgID, = struct.unpack_from('<b', data, bmgIDsTableOffset + i)
                if bmgID != 0 or index != 0:
                    self.labels.append((bmgID, index))

        self.scripts = []
        def parseFLI1(offset, length):
            count, entryLength, unk0C = struct.unpack_from('<HHI', data, offset + 8)
            assert entryLength == 8
            # unk0C is always 0, as far as I can tell

            for i in range(count):
                id, index = struct.unpack_from('<IHxx', data, offset + 16 + i * 8)
                self.scripts.append((id, index))

        offset = 0x20
        for i in range(sectionCount):
            sectionMagic, sectionLen = struct.unpack_from('<4sI', data, offset)
            if sectionMagic == b'INF1':
                parseINF1(offset, sectionLen)
            elif sectionMagic == b'DAT1':
                parseDAT1(offset, sectionLen)
            elif sectionMagic == b'FLW1':
                parseFLW1(offset, sectionLen)
            elif sectionMagic == b'FLI1':
                parseFLI1(offset, sectionLen)
            else:
                raise ValueError('Unknown BMG section: ' + repr(sectionMagic))
            offset += sectionLen

        # Now we just need to read the messages.

        self.messages = []
        for offset, attribs in INF1:

            # The "currentStringStart" setup may seem needlessly
            # confusing, but it's intended to keep the number of calls
            # to bytes.decode() to a minimum. Based on my testing, it
            # really does make the code run significantly faster.

            stringParts = []
            currentStringStart = offset

            nextBytes = DAT1[offset : offset + 2]
            while nextBytes != b'\0\0':
                if nextBytes == b'\x1A\x00': # escape sequence
                    if currentStringStart and currentStringStart != offset:
                        stringParts.append(DAT1[currentStringStart:offset].decode('utf-16le'))
                    escapeLen, escapeType = DAT1[offset + 2 : offset + 4]
                    escapeData = DAT1[offset + 4 : offset + escapeLen]
                    stringParts.append(Message.Escape(escapeType, escapeData))
                    offset += escapeLen
                    currentStringStart = offset
                else:
                    offset += 2

                nextBytes = DAT1[offset : offset + 2]

            if currentStringStart and currentStringStart != offset:
                stringParts.append(DAT1[currentStringStart:offset].decode('utf-16le'))

            self.messages.append(Message(attribs, stringParts, offset == 0))


    @classmethod
    def fromMessages(cls, messages,
                     instructions=None, labels=None, scripts=None,
                     *, id=0):
        """
        Create a BMG from a list of messages.
        """
        self = cls(id=id)
        self.messages = messages

        if instructions is not None:
            self.instructions = instructions
        if labels is not None:
            self.labels = labels
        if scripts is not None:
            self.scripts = scripts

        return self


    @classmethod
    def fromFile(cls, filePath, *args, **kwargs):
        """
        Load a BMG from a filesystem file.
        """
        with open(filePath, 'rb') as f:
            return cls(f.read(), *args, **kwargs)


    def save(self):
        """
        Generate file data representing this BMG.
        """
        data = bytearray(0x20)

        instructionsCount = len(self.instructions)
        if instructionsCount % 2: instructionsCount += 1

        labelsCount = len(self.labels)
        while labelsCount % 8: labelsCount += 1

        numSections = 2
        if self.instructions: numSections += 1
        if self.scripts: numSections += 1

        INF1 = bytearray(16)
        DAT1 = bytearray(10) # Starts with 2 null bytes for whatever reason
        FLW1 = bytearray(16)
        FLI1 = bytearray(16)

        for message in self.messages:
            offset = 0 if message.isNull else len(DAT1) - 8
            INF1.extend(struct.pack('<II', offset, message.info))
            if not message.isNull:
                DAT1.extend(message.save())

        for inst in self.instructions:
            if hasattr(inst, 'save'):
                inst = inst.save()
            if len(inst) != 8:
                raise ValueError(f'Length of instruction {inst} is not 8 bytes!')
            FLW1.extend(inst)
        while len(FLW1) % 16: FLW1.extend(b'\0' * 8)

        for bmgID, instIndex in self.labels:
            FLW1.extend(struct.pack('<h', instIndex))
        for _ in range(labelsCount - len(self.labels)):
            FLW1.extend(b'\0\0')
        for bmgID, instIndex in self.labels:
            FLW1.extend(struct.pack('<b', bmgID))

        for id, startIndex in self.scripts:
            FLI1.extend(struct.pack('<II', id, startIndex))

        # Sections' lengths must be 32-byte aligned
        while len(INF1) % 32: INF1.append(0)
        while len(DAT1) % 32: DAT1.append(0)
        while len(FLW1) % 32: FLW1.append(0)

        # FLI1's length isn't actually padded, but the length it claims
        # in its header is. (I know. It's confusing.)
        FLI1len = len(FLI1)
        while FLI1len % 32: FLI1len += 1

        # Pack section headers
        struct.pack_into('<4sIHHI', INF1, 0,
            b'INF1', len(INF1), len(self.messages), 8, self.id)
        struct.pack_into('<4sI', DAT1, 0, b'DAT1', len(DAT1))
        if self.instructions:
            struct.pack_into('<4sIHH', FLW1, 0,
                b'FLW1', len(FLW1), instructionsCount, labelsCount)
        if self.scripts:
            struct.pack_into('<4sIHH', FLI1, 0,
                b'FLI1', FLI1len, len(self.scripts), 8)

        # Insert the sections
        data.extend(INF1)
        data.extend(DAT1)
        if self.instructions: data.extend(FLW1)
        if self.scripts: data.extend(FLI1)

        # Pack the BMG header
        totalLen = len(data)
        while totalLen % 32: totalLen += 1
        struct.pack_into('<8s3I', data, 0, b'MESGbmg1', totalLen, numSections, self.unk10)

        return bytes(data)


    def saveToFile(self, filePath):
        """
        Generate file data representing this BMG, and save it to a
        filesystem file.
        """
        d = self.save()
        with open(filePath, 'wb') as f:
            f.write(d)


    def __str__(self):
        return (f'<bmg id={self.id} '
                f'({len(self.messages)} messages,'
                f'{len(self.scripts)} scripts)>')


    def __repr__(self):
        optionalArgs = []

        if self.instructions:
            optionalArgs.append(repr(self.instructions))

        if self.labels:
            p = ''
            if len(optionalArgs) < 1:
                p = 'labels='
            optionalArgs.append(p + repr(self.labels))

        if self.scripts:
            p = ''
            if len(optionalArgs) < 2:
                p = 'scripts='
            optionalArgs.append(p + repr(self.scripts))

        optionalArgs.append(f'id={self.id:#x}')

        return (f'{type(self).__name__}.fromMessages('
                f'{self.messages!r}{", ".join(optionalArgs)})')


class Message:
    """
    A single message in a BMG file.
    """

    class Escape:
        """
        An escape sequence within a BMG message. Escape sequences have a
        type and optional parameter data.
        """

        def __init__(self, type=0, data=b''):
            self.type = type
            self.data = data

            # Type = 4 is used for pluralization
            # (and there's a couple parameters -- need to look into that)

        def save(self):
            """
            Generate binary data representing this escape sequence.
            """
            data = bytearray(b'\x1A\x00')
            data.append(len(self.data) + 4)
            data.append(self.type)
            data.extend(self.data)
            return data

        def __repr__(self):
            return f'{type(self).__name__}({self.type!r}, {self.data!r})'

        def __str__(self):
            return f'[{self.type}:{self.data.hex()}]'

    info = 0
    stringParts = []
    isNull = False

    def __init__(self, info=0, stringParts=None, isNull=False):
        self.info = info
        self.stringParts = [] if stringParts is None else stringParts
        self.isNull = isNull

    def save(self):
        """
        Generate binary data representing this message.
        """
        data = bytearray()
        for part in self.stringParts:
            if isinstance(part, str):
                data.extend(part.encode('utf-16le'))
            else:
                data.extend(part.save())
        data.extend(b'\0\0')
        return data

    def __repr__(self):
        return f'{type(self).__name__}({hex(self.info)}, {self.stringParts!r})'

    def __str__(self):
        return ''.join(str(s) for s in self.stringParts)



# Copyright 2019 RoadrunnerWMC
#
# This file is part of ndspy.
#
# ndspy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# ndspy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with ndspy.  If not, see <https://www.gnu.org/licenses/>.
"""
Support for sound groups within SDATs.
"""


import enum


class GroupEntryType(enum.IntEnum):
    """
    An enumeration that distinguishes between the types of files that
    a group entry can refer to.
    """
    SSEQ = 0
    SBNK = 1
    SWAR = 2
    SSAR = 3


class SWARLoadMethod(enum.IntEnum):
    """
    An enumeration that distinguishes between the ways in which a *SWAR*
    can be loaded.
    """
    FILE_IDS = 0
    SWAR_IDS = 1


class GroupEntry:
    """
    An entry in a sound group.
    """

    # Default options value is "2" to set SWAR load method to SWAR IDs,
    # which is what we want to be the default
    def __init__(self, type=GroupEntryType.SSEQ, options=2, id=0):
        self.type = GroupEntryType(type)
        self.id = id

        # The variable names are based on actual code RE.

        # Note that a SSEQ will pass its "options" value on to the
        # SBNK it loads, and a SBNK will do the same with the SWARs it
        # loads.

        self.loadSSEQ = bool(options & 1)
        self.loadSBNKSWARsFrom = SWARLoadMethod(bool(options & 2))
        self.loadSWAR = bool(options & 4)
        self.loadSSAR = bool(options & 8)


    @classmethod
    def fromFlags(cls, type, id, loadSSEQ=False,
            loadSBNKSWARsFrom=SWARLoadMethod.SWAR_IDS, loadSWAR=False,
            loadSSAR=False):
        """
        Create a sound group entry from individual attribute values.
        """

        self = cls(type, 0, id)

        self.loadSSEQ = loadSSEQ
        self.loadSBNKSWARsFrom = loadSBNKSWARsFrom
        self.loadSWAR = loadSWAR
        self.loadSSAR = loadSSAR

        return self


    def save(self):
        """
        Return this sound group entry's type, options value, and ID as a
        3-tuple. This matches the parameters of the default class
        constructor.
        """

        options = 0
        if self.loadSSEQ: options |= 1
        if self.loadSBNKSWARsFrom == SWARLoadMethod.SWAR_IDS: options |= 2
        if self.loadSWAR: options |= 4
        if self.loadSSAR: options |= 8

        return self.type, options, self.id


    def __str__(self):
        typeName = {
            GroupEntryType.SSEQ: 'sseq',
            GroupEntryType.SBNK: 'sbnk',
            GroupEntryType.SWAR: 'swar',
            GroupEntryType.SSAR: 'ssar',
            }.get(self.type, str(int(self.type)))

        optionsL = []
        if self.loadSSEQ: optionsL.append('load-sseq')
        if self.loadSBNKSWARsFrom:
            method = {
                SWARLoadMethod.FILE_IDS: 'file-id',
                SWARLoadMethod.SWAR_IDS: 'swar-id',
                }[self.loadSBNKSWARsFrom]
            optionsL.append('load-swars-by-' + method)
        if self.loadSWAR: optionsL.append('load-swar')
        if self.loadSSAR: optionsL.append('load-ssar')
        options = ' '.join(optionsL)

        return f'<group-entry-{typeName} {self.id}{" " if options else ""}{options}>'


    def __repr__(self):
        args = []

        args.append(repr(self.type))
        args.append(repr(self.id))

        if self.loadSSEQ:
            args.append(f'loadSSEQ={self.loadSSEQ!r}')
        if self.loadSBNKSWARsFrom != SWARLoadMethod.SWAR_IDS:
            args.append(f'loadSBNKSWARsFrom={self.loadSBNKSWARsFrom!r}')
        if self.loadSWAR:
            args.append(f'loadSWAR={self.loadSWAR!r}')
        if self.loadSSAR:
            args.append(f'loadSSAR={self.loadSSAR!r}')

        return f'{type(self).__name__}.fromFlags({", ".join(args)})'




import contextlib

from .. import _common
from .. import soundBank
from .. import soundSequence


_MONO = soundSequence.MonoPolySequenceEvent.Value.MONO
_POLY = soundSequence.MonoPolySequenceEvent.Value.POLY


def minifyByEvents(eventsIter, sbnk, swars=None):
    """
    Minify the given SBNK and optionally SWARs.

    This is a low-level minification function; before you consider using
    it, check whether you can instead use a higher-level variation:

    - minifyBySseq()
    - minifyBySseqs()
    - SSEQMusic.minify()

    eventsIter is an iterator over all events that use the SBNK. This
    list must be exhaustive; if not, SSEQs you omit won't be updated to
    reflect changes in instrument IDs, which will result in broken
    music. Also, too many instruments might be deleted.

    sbnk is the SBNK to be minified.

    swars is the list of (up to four) SWARs that the SBNK uses. If
    provided, anything within them that the SBNK doesn't use (after it's
    been minified) will be removed. This is optional.

    Note that minifying SWARs is usually the most important part. (SBNKs
    are generally tiny in comparison.)

    Doesn't return anything; the objects you pass as params are
    modified.

    Warning: you should only use this after you're completely done
    editing your sseq (programmatically or otherwise).

    Warning 2: this is a one-way operation that strips away all
    instruments you're not using (and even unused aspects of the
    instruments you do use). Don't delete the original after you use
    this, because it's not very easy to undo this.

    Note to whoever's cleaning up this documentation: those warnings
    should be copypasted (and modified as needed) to the other minify
    functions that build on this one.

    .. todo::
        Another optimization that can be done is combining consecutive
        rests. You have to check that they're not a jump target of some
        sort first, though! (including implicit jumps defined by track
        start events and the like)

        Also, something else that's a bit easier is removing instruments
        that are emptied of all of their notes (e.g. they're referenced
        but not actually used).

        Also, while you can't indiscriminately delete raw-data events,
        you can check for any that immediately follow a (non-conditional)
        end-of-track or jump event and are not a jump target, and delete
        those. This covers the very common case of needlessly ending a
        track with 0xFFFF.

        Also, each of these optimizations should really be its own
        function, since this code is getting kind of complicated.
    """

    # This used to be a parameter defaulting to False, since it was
    # unreliable and could break songs; however, it's stable enough now
    # that it shouldn't be publicly exposed. It'll still be here as a
    # constant for debugging purposes, though.
    MINIFY_INSTRUMENTS=True

    events = list(eventsIter)

    # Remove unused entries in SBNK relative to SSEQ
    usedInsts = set()
    playedNotes = {}
    currentInstrument = None
    for e in events:
        if isinstance(e, soundSequence.InstrumentSwitchSequenceEvent):
            currentInstrument = e.instrumentID
            usedInsts.add(currentInstrument)
            if currentInstrument not in playedNotes:
                playedNotes[currentInstrument] = set()
        elif isinstance(e, soundSequence.NoteSequenceEvent):
            if currentInstrument is not None:
                playedNotes[currentInstrument].add(e.type)
    usedInsts = sorted(usedInsts)
    sbnk.instruments = [sbnk.instruments[num] for num in usedInsts]
    playedNotesPerInst = [playedNotes[num] for num in usedInsts]
    for e in events:
        if isinstance(e, soundSequence.InstrumentSwitchSequenceEvent):
            e.instrumentID = usedInsts.index(e.instrumentID)

    # (and instruments that were useless in the first place)
    sbnk.inaccessibleInstruments.clear()

    # Remove unused entries in SWAR relative to SBNK
    usedWaves = set()
    for inst, notes in zip(sbnk.instruments, playedNotesPerInst):
        if isinstance(inst, soundBank.SingleNoteInstrument):
            usedWaves.add((inst.noteDefinition.waveArchiveIDID, inst.noteDefinition.waveID))
        elif isinstance(inst, soundBank.RangeInstrument):
            noteNum = inst.firstPitch - 1
            for d in inst.noteDefinitions:
                noteNum += 1
                if MINIFY_INSTRUMENTS and (noteNum not in notes):
                    continue
                usedWaves.add((d.waveArchiveIDID, d.waveID))
        elif isinstance(inst, soundBank.RegionalInstrument):
            low = 0
            for reg in inst.regions:
                if MINIFY_INSTRUMENTS:
                    # Try to find the note; if it can't be found,
                    # skip this swav
                    for i in range(low, reg.lastPitch + 1):
                        if i in notes:
                            break
                    else:
                        continue
                usedWaves.add((reg.noteDefinition.waveArchiveIDID, reg.noteDefinition.waveID))
                low = reg.lastPitch + 1
    usedWaves = sorted(usedWaves, key=lambda e: e[0] * 99999999 + e[1])

    for swarNum, swar in enumerate(swars):
        newWaves = []
        for (swarNum2, swavNum) in usedWaves:
            if swarNum != swarNum2: continue
            newWaves.append(swar.waves[swavNum])
        swar.waves = newWaves

    for inst in sbnk.instruments:
        if isinstance(inst, soundBank.SingleNoteInstrument):
            inst.noteDefinition.waveID = usedWaves.index(
                (inst.noteDefinition.waveArchiveIDID, inst.noteDefinition.waveID))
        elif isinstance(inst, soundBank.RangeInstrument):
            filteredDefinitions = []
            dummyDefinition = inst.noteDefinitions[0]
            for d in inst.noteDefinitions:
                if (d.waveArchiveIDID, d.waveID) in usedWaves:
                    d.waveID = usedWaves.index((d.waveArchiveIDID, d.waveID))
                    filteredDefinitions.append(d)
                else:
                    filteredDefinitions.append(dummyDefinition)
            inst.noteDefinitions = filteredDefinitions
        elif isinstance(inst, soundBank.RegionalInstrument):
            filteredRegions = []
            for reg in inst.regions:
                if (reg.noteDefinition.waveArchiveIDID, reg.noteDefinition.waveID) in usedWaves:
                    reg.noteDefinition.waveID = usedWaves.index(
                        (reg.noteDefinition.waveArchiveIDID, reg.noteDefinition.waveID))
                    filteredRegions.append(
                        soundBank.RegionalInstrument.Region(
                            reg.lastPitch, reg.noteDefinition))
            inst.regions = filteredRegions
            if inst.regions:
                inst.regions[-1].lastPitch = 127 # required


def minifyBySseq(sseq, sbnk, swars=None):
    """
    Minify the given SBNK and optionally SWARs.

    This function is ideal for an SBNK that's used by only one SSEQ.
    This is a common case. Note (todo: figure out how to word this
    better): don't just think, like, "I have three songs using HGSS, so
    this function isn't for me." In most cases, what you actually have
    are three songs using three different copies of HGSS.
    """
    minifyByEvents(sseq.events, sbnk, swars)


def minifyBySseqs(sseqs, sbnk, swars=None):
    """
    Minify the given SBNK and optionally SWARs.

    This function is ideal for an SBNK that's only used by SSEQs. If
    your SBNK is only used by one SSEQ, you can use minifyBySseq()
    instead. That's equivalent to minifyBySseqs([sseq], sbnk, swars).
    """
    def events():
        for sseq in sseqs:
            sseq.parse()
            yield from sseq.events
    minifyByEvents(events(), sbnk, swars)


def durationOf(events,
        initialMonoPolyMode=_MONO):
    """
    Return the sum of all times taken by the given list of events.
    This does NOT take into account any additional time taken by jumps
    or such.
    """

    monoPolyMode = initialMonoPolyMode

    def durationOfEvent(event):
        nonlocal monoPolyMode
        if isinstance(event, soundSequence.RestSequenceEvent):
            return event.duration
        elif (isinstance(event, soundSequence.NoteSequenceEvent)
                and monoPolyMode is _MONO):
            return event.duration
        elif isinstance(event, soundSequence.MonoPolySequenceEvent):
            monoPolyMode = event.value
        return 0

    return sum(map(durationOfEvent, events))


def moveInTime(events, startIndex, time,
        initialMonoPolyMode=_MONO,
        cut=False, append=False):
    """
    Given a list of events and a starting event index, move in time by
    AT MOST the requested amount. Return, as a 3-tuple:
        - the index of the first event at this time (or the length of
          the list if applicable)
        - the amount of time that was moved
        - the mono/poly mode at this time
    Zero-duration events at the end of the time period moved through
    will not be consumed.

    `monoPolyMode` is the initial mono/poly mode. It defaults to `Mono`
    because the DS does. If you're starting at the beginning of a track,
    you shouldn't have to worry about this.

    If `cut` is True, a rest sequence event may be cut into two in order
    to move forward the exact requested amount.

    If `append` is True, a rest sequence event may be appended to a
    track in order to move forward the exact requested amount. (This
    will not occur for looped tracks.)

    Thus, if both of these are False, the events list will be guaranteed
    to be unmodified after the call.

    .. warning::
        Moving backwards in time is currently unsupported, and currently
        throws a ValueError. This may become supported in a future
        release.
    """

    # Special cases for moving backwards in time, and moving zero time
    if time < 0:
        raise ValueError(f'Moving backwards in time is currently unsupported (time={time})')
    elif time == 0:
        return startIndex, 0, initialMonoPolyMode

    elapsedTime = 0
    target = time

    if startIndex > len(events):
        raise ValueError('moveInTime(): startIndex is greater than the'
                         ' list of events!')
    idx = startIndex

    monoPolyMode = initialMonoPolyMode

    time = 0
    while idx < len(events):
        e = events[idx]
        d = durationOf([e], monoPolyMode)
        newElapsedTime = elapsedTime + d

        if isinstance(e, soundSequence.MonoPolySequenceEvent):
            monoPolyMode = e.value

        if newElapsedTime == target:
            # We've hit it exactly! This is where we stop.
            return idx + 1, newElapsedTime, monoPolyMode

        elif newElapsedTime >= target:
            # This event would put us past the target time.
            # Split it in half if it's a rest and we're allowed to;
            # otherwise, don't.
            if isinstance(e, soundSequence.RestSequenceEvent) and cut:
                firstHalfLength = target - elapsedTime
                firstHalf = e
                fullRestLength = firstHalf.duration
                firstHalf.duration = firstHalfLength
                secondHalf = soundSequence.RestSequenceEvent(
                    fullRestLength - firstHalfLength)
                events.insert(idx + 1, secondHalf)
                return idx + 1, target, monoPolyMode

            else:
                return idx, elapsedTime, monoPolyMode

            break

        elapsedTime += d
        idx += 1

    # We made it all the way through without reaching the desired time.
    # Now we extend the list if the caller requested that...
    if append:
        newRest = soundSequence.RestSequenceEvent(target - elapsedTime)
        events.append(newRest)
        elapsedTime = target

    # ... and, whether we extended the list or not, we now return the
    # end of the list.
    return len(events), elapsedTime, monoPolyMode


def moveTo(events, start, end,
        monoPolyMode=_MONO):
    """
    Like moveInTime(), but you move to a specific event. Useful for
    checking the mono/poly mode at that point or the time elapsed
    between two events.
    Returns the elapsed time and the mono/poly mode.

    TODO: do we actually need this now that durationOf() exists?
    """
    raise NotImplementedError


@contextlib.contextmanager
def SimultaneousEventsModifier(events):
    """
    A context manager over a list of events, where you can make changes
    to the list (via the context manager) and the list itself won't
    change until the context is exited. Lets you do nice things like
    events = [a, b, c]
    with SimultaneousEventsModifier(events) as mod:
        events.insert(1, d)
        events.insert(1, e)
        events.insert(1, f)
        events.insert(2, g)
    print(events)
    [a, d, e, f, b, g, c]
    since the indices are always relative to the original list!

    .. warning::
        Don't use this with functions that modify the events list and
        expect you to work with the modified list (e.g. moveInTime with
        cut = True)! Then your indices will be off.

    TODO: probably make this into a legit class instead of this contextmanager
    """
    class LogCreator(list):
        """
        A subclass of list which overrides functions that would modify
        the list, and just writes them to a log instead (.log)
        """
        def __init__(self, other=None):
            if other is None: other = []
            super().__init__(other)
            self.log = {}

        def insert(self, index, item):
            if index not in self.log:
                self.log[index] = {}
            if 'additions' not in self.log[index]:
                self.log[index]['additions'] = []
            self.log[index]['additions'].append(item)

        def __setitem__(self, index, item):
            if index not in self.log:
                self.log[index] = {}
            self.log[index]['replace'] = item

        def __delitem__(self, index):
            if index not in self.log:
                self.log[index] = {}
            self.log[index]['delete'] = True


    logger = LogCreator(events)
    yield logger

    log = logger.log
    for idx in reversed(sorted(log)):
        # If a replacement and a deletion are both requested, just do
        # the replacement
        if 'replace' in log[idx]:
            events[idx] = log[idx]['replace']
        elif log[idx].get('delete', False):
            del events[idx]

        for event in reversed(log[idx].get('additions', [])):
            events.insert(idx, event)


class SSEQMusic:
    """
    An opinionated class that provides a higher-level API around
    SSEQ-based music.

    This allows you to perform some higher-level operations on its
    contents. Unlike the SSEQ class, this assumes some things about the
    SSEQ's structure in order to make it easier to edit.

    .sseq/.sbnk/.swars are NOT guaranteed to be kept up to date until
    just after .save() is called. Also, sbnk and swars are explicitly
    allowed to be None. (Only functions that literally cannot function
    without them, such as Minify, should fail. Also, Minify should be
    able to minify just the SSEQ and leave the SBNK alone if the SSEQ
    and SBNK but not SWAR are present.)
    """

    class Track:
        """
        Class that represents a channel in a SSEQ file.
        """
        def __init__(self, events):
            self.events = list(events)
            if self.events:
                if isinstance(self.events[-1], soundSequence.EndTrackSequenceEvent):
                    self.events.pop()


        @property
        def looped(self):
            return any(isinstance(e, soundSequence.JumpSequenceEvent)
                for e in self.events)


        def loop(self, start, end, *, clearTrailing=True):
            """
            Loop the track.
            `start` and `end` are integers, in the same units used for
            durations in the SSEQ.
            If `clearTrailing` is set, all events after the end of the
            loop will be deleted.
            """
            if self.looped:
                raise RuntimeError("Can't loop a track that's already"
                                   ' looped')

            loopStartIdx, actualStartTime, _ = moveInTime(
                self.events, 0, start, cut=True, append=True)

            if actualStartTime != start:
                raise RuntimeError(f'Unable to start the loop at {start}'
                                   f' (only able to start it at'
                                   f' {actualStartTime})')

            loopEndIdx, actualEndTime, _ = moveInTime(
                self.events, 0, end, cut=True, append=True)

            if actualEndTime != end:
                raise RuntimeError(f'Unable to end the loop at {end}'
                                   f' (only able to end it at'
                                   f' {actualEndTime})')

            jumpTarget = self.events[loopStartIdx]
            jump = soundSequence.JumpSequenceEvent(jumpTarget)
            self.events.insert(loopEndIdx, jump)

            if clearTrailing:
                self.events = self.events[:loopEndIdx + 1]


        def _modifyNoteVelocities(self, func, *, clamp=True):
            for e in self.events:
                if not isinstance(e, soundSequence.NoteSequenceEvent):
                    continue
                newVelocity = int(func(e.velocity))
                if not (0 <= newVelocity <= 127):
                    if clamp:
                        newVelocity = min(max(newVelocity, 0), 127)
                    else:
                        raise ValueError('Attempted to set note velocity to'
                                        f' {newVelocity} (not in range'
                                         ' 0..127)')
                e.velocity = newVelocity


        def addToNoteVelocities(self, amount, *, clamp=True):
            def adder(velocity):
                return velocity + amount
            self._modifyNoteVelocities(adder, clamp=clamp)


        def multiplyNoteVelocitiesBy(self, factor, *, clamp=True):
            def multiplier(velocity):
                return velocity * factor
            self._modifyNoteVelocities(multiplier, clamp=clamp)


        # Properties that add/remove events at the beginning of the list:
        # initialTempo, initialMonoPoly, initialChannelVolume, initialInstrument, etc

        def _findControlEventAtBeginning(self, type):
            for e in self.events:
                if isinstance(e, type):
                    return e
                elif (isinstance(e, soundSequence.NoteSequenceEvent)
                        or isinstance(e, soundSequence.RestSequenceEvent)):
                    break

        def _handleInitialValueGet(self, type, *, isInstrumentSwitch=False):
            event = self._findControlEventAtBeginning(type)
            if event is None: return None
            if isInstrumentSwitch:
                return event.bankID, event.instrumentID
            else:
                return event.value

        def _handleInitialValueSet(self, type, value, *, isInstrumentSwitch=False):

            event = self._findControlEventAtBeginning(type)

            if event is None and value is None:
                # No event exists yet and we're supposed to delete it.
                # Do nothing.
                pass

            elif event is None:
                # No event exists yet and we're supposed to give it a
                # new value. Make a new event and add it.
                if isInstrumentSwitch:
                    event = type(value[0], value[1])
                else:
                    event = type(value)
                self.events.insert(0, event)

            elif value is None:
                # An event exists and we're supposed to delete it.
                self.events.remove(event)

            else:
                # An event exists and we're supposed to give it a new
                # value.
                if isInstrumentSwitch:
                    event.bankID = value[0]
                    event.instrumentID = value[1]
                else:
                    event.value = value

        @property
        def initialInstrument(self):
            return self._handleInitialValueGet(
                soundSequence.InstrumentSwitchSequenceEvent,
                isInstrumentSwitch=True)
        @initialInstrument.setter
        def initialInstrument(self, value):
            return self._handleInitialValueSet(
                soundSequence.InstrumentSwitchSequenceEvent,
                value,
                isInstrumentSwitch=True)

        @property
        def initialPan(self):
            return self._handleInitialValueGet(soundSequence.PanSequenceEvent)
        @initialPan.setter
        def initialPan(self, value):
            return self._handleInitialValueSet(soundSequence.PanSequenceEvent,
                                               value)

        @property
        def initialVolume(self):
            return self._handleInitialValueGet(
                soundSequence.TrackVolumeSequenceEvent)
        @initialVolume.setter
        def initialVolume(self, value):
            return self._handleInitialValueSet(
                soundSequence.TrackVolumeSequenceEvent, value)

        @property
        def initialTranspose(self):
            return self._handleInitialValueGet(
                soundSequence.TransposeSequenceEvent)
        @initialTranspose.setter
        def initialTranspose(self, value):
            return self._handleInitialValueSet(
                soundSequence.TransposeSequenceEvent, value)

        @property
        def initialPriority(self):
            return self._handleInitialValueGet(
                soundSequence.TrackPrioritySequenceEvent)
        @initialPriority.setter
        def initialPriority(self, value):
            return self._handleInitialValueSet(
                soundSequence.TrackPrioritySequenceEvent, value)

        @property
        def initialMonoPoly(self):
            return self._handleInitialValueGet(
                soundSequence.MonoPolySequenceEvent)
        @initialMonoPoly.setter
        def initialMonoPoly(self, value):
            return self._handleInitialValueSet(
                soundSequence.MonoPolySequenceEvent, value)

        @property
        def initialAttackRate(self):
            return self._handleInitialValueGet(
                soundSequence.AttackRateSequenceEvent)
        @initialAttackRate.setter
        def initialAttackRate(self, value):
            return self._handleInitialValueSet(
                soundSequence.AttackRateSequenceEvent, value)

        @property
        def initialDecayRate(self):
            return self._handleInitialValueGet(
                soundSequence.DecayRateSequenceEvent)
        @initialDecayRate.setter
        def initialDecayRate(self, value):
            return self._handleInitialValueSet(
                soundSequence.DecayRateSequenceEvent, value)

        @property
        def initialSustainRate(self):
            return self._handleInitialValueGet(
                soundSequence.SustainRateSequenceEvent)
        @initialSustainRate.setter
        def initialSustainRate(self, value):
            return self._handleInitialValueSet(
                soundSequence.SustainRateSequenceEvent, value)

        @property
        def initialReleaseRate(self):
            return self._handleInitialValueGet(
                soundSequence.ReleaseRateSequenceEvent)
        @initialReleaseRate.setter
        def initialReleaseRate(self, value):
            return self._handleInitialValueSet(
                soundSequence.ReleaseRateSequenceEvent, value)

        @property
        def initialTempo(self):
            return self._handleInitialValueGet(
                soundSequence.TempoSequenceEvent)
        @initialTempo.setter
        def initialTempo(self, value):
            return self._handleInitialValueSet(
                soundSequence.TempoSequenceEvent, value)


        def __str__(self):
            linesList = ['<track']
            linesList.append(soundSequence.printSequenceEventList(
                self.events, {}, ' ' * 2))
            linesList.append('>')
            return '\n'.join(linesList)


        def __repr__(self):
            return f'{type(self).__name__}({self.events!r})'


    def __init__(self, sseq, sbnk=None, swars=None):
        self.sseq = sseq
        self.sbnk = sbnk
        self.swars = swars
        if not _common.isIterable(self.swars):
            raise TypeError('SWAR list is not iterable (did you'
                ' accidentally pass a SWAR to the constructor instead'
                ' of a list of SWARs?)')

        # Make a dictionary of tracks
        # Each track's events are considered to end whenever another
        # track begins.
        self.tracks = {}
        if isinstance(sseq.events[0],
                soundSequence.DefineTracksSequenceEvent):

            # Find events that begin tracks (so that we know where to
            # end tracks)
            # (the beginning of track 0 is omitted from this list, but
            # that's OK)
            trackStarts = []
            i = 1
            while isinstance(sseq.events[i],
                    soundSequence.BeginTrackSequenceEvent):
                trackStarts.append(sseq.events[i].firstEvent)
                i += 1

            # Track 0 doesn't have a BeginTrack event
            if 0 in sseq.events[0].trackNumbers:
                # Find where the BeginTrack events end
                startIdx = 1
                while isinstance(sseq.events[startIdx],
                        soundSequence.BeginTrackSequenceEvent):
                    startIdx += 1
                endIdx = startIdx + 1
                while (endIdx < len(sseq.events)
                        and sseq.events[endIdx] not in trackStarts):
                    endIdx += 1

                trackEvents = sseq.events[startIdx : endIdx]
                self.tracks[0] = self.Track(trackEvents)

            # Create the Track objects
            i = 1
            while isinstance(sseq.events[i],
                    soundSequence.BeginTrackSequenceEvent):
                e = sseq.events[i]
                startIdx = sseq.events.index(e.firstEvent)
                endIdx = startIdx + 1
                while (endIdx < len(sseq.events)
                        and sseq.events[endIdx] not in trackStarts):
                    endIdx += 1

                trackEvents = sseq.events[startIdx : endIdx]
                self.tracks[e.trackNumber] = self.Track(trackEvents)
                i += 1

        else:
            self.tracks[0] = self.Track(sseq.events)


    def loop(self, start, end, *, clearTrailing=True):
        """
        Loop all tracks in the SSEQ.
        `start` and `end` are integers, in the same units used for
        durations in the SSEQ.
        `clearTrailing` has the same meaning as in Track.loop().
        """
        for track in self.tracks.values():
            track.loop(start, end, clearTrailing=clearTrailing)


    def minify(self):
        """
        Minify the song (SSEQ, optionally SBNK, optionally SWAR)
        The SBNK and SWAR will be modified, so be forewarned
        (this may cause unintended side-effects if you're reusing the
        same bank/swar in multiple places)
        Also, don't use this if multiple songs use the same SBNK.
        Use minifyBySseqs() instead in that case, after calling save()
        on this object.
        """
        if self.sbnk is None:
            # There's nothing to minify if there's no SBNK
            return

        def events():
            for track in self.tracks.values():
                yield from track.events
        minifyByEvents(events(), self.sbnk, self.swars)


    def collapseTrackIDs(self):
        """
        If the current track IDs are like
        0, 1, 2, 4, 5, 8, 9, 10
        , this function renumbers them to the saner
        0, 1, 2, 3, 4, 5, 6, 7
        while preserving their order.
        """
        newTracks = {}
        for i, id in enumerate(sorted(self.tracks.keys())):
            newTracks[i] = self.tracks[id]
        self.tracks = newTracks


    def save(self):
        """
        Update self.sseq / self.sbnk / self.swars to reflect changes
        made in self.tracks, and return them. You can safely use those
        properties after calling this function, until you start making
        changes again.
        """
        # Rebuild sseq.events
        if len(self.tracks) == 1 and 0 in self.tracks:
            # There's only one track and it's track 0. Don't bother
            # defining tracks and stuff.
            self.sseq.events = list(self.tracks[0].events)
            if not (isinstance(self.sseq.events[-1],
                               soundSequence.EndTrackSequenceEvent)
                    or isinstance(self.sseq.events[-1],
                                  soundSequence.JumpSequenceEvent)):
                self.sseq.events.append(soundSequence.EndTrackSequenceEvent())

        else:
            beginning = []
            contents = []

            dfse = soundSequence.DefineTracksSequenceEvent(set(self.tracks))
            beginning.append(dfse)

            for trackNum in sorted(self.tracks):
                track = self.tracks[trackNum]
                if not track.events:
                    # Um.
                    continue

                if trackNum != 0:
                    beginning.append(
                        soundSequence.BeginTrackSequenceEvent(
                            trackNum, track.events[0]))
                contents.extend(track.events)
                if not (isinstance(contents[-1],
                                   soundSequence.EndTrackSequenceEvent)
                        or isinstance(contents[-1],
                                      soundSequence.JumpSequenceEvent)):
                    contents.append(soundSequence.EndTrackSequenceEvent())

            self.sseq.events = beginning + contents

        return self.sseq, self.sbnk, self.swars


    def __str__(self):
        trackNums = sorted(self.tracks)
        trackNumStrs = [str(x) for x in trackNums]
        return f'<sseq music [tracks: {", ".join(trackNumStrs)}]>'


    def __repr__(self):
        return f'{type(self).__name__}({self.sseq!r})'



# Copyright 2019 RoadrunnerWMC
#
# This file is part of ndspy.
#
# ndspy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# ndspy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with ndspy.  If not, see <https://www.gnu.org/licenses/>.
"""
Support for sequence players.
"""

class SequencePlayer:
    """
    A sequence player.
    """
    def __init__(self, maxSequences=1, channels=None, heapSize=0):
        self.maxSequences = maxSequences
        self.channels = channels if channels is not None else set()
        self.heapSize = heapSize


    def save(self):
        """
        Return this sequence player's .maxSequences, .channels, and
        .heapSize as a triple. This matches the parameters of the
        default class constructor.
        """
        return (self.maxSequences, self.channels, self.heapSize)


    def __str__(self):
        c = str(sorted(self.channels))
        return f'<sequence-player maxSequences={self.maxSequences} channels={c} heapSize={self.heapSize}>'


    def __repr__(self):
        return f'{type(self).__name__}({self.maxSequences!r}, {self.channels!r}, {self.heapSize!r})'



# Copyright 2019 RoadrunnerWMC
#
# This file is part of ndspy.
#
# ndspy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# ndspy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with ndspy.  If not, see <https://www.gnu.org/licenses/>.
"""
Support for stream players.
"""


class StreamPlayer:
    """
    A stream player.
    """
    def __init__(self, channels=None):
        self.channels = channels if channels is not None else []


    def save(self):
        """
        Return this sequence player's .channels attribute. This matches
        the parameter of the default class constructor.
        """
        return list(self.channels)


    def __str__(self):
        return f'<stream-player {self.channels}>'


    def __repr__(self):
        if self.channels:
            return f'{type(self).__name__}()'
        else:
            return f'{type(self).__name__}({self.channels!r})'



# Copyright 2019 RoadrunnerWMC
#
# This file is part of ndspy.
#
# ndspy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# ndspy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with ndspy.  If not, see <https://www.gnu.org/licenses/>.

from . import _common


# These are populated at the bottom of the file.
LUT_UNPACKED = [None] * 0x10000
LUT_UNPACKED_255 = [None] * 0x10000
LUT_RGBA_255 = [None] * 0x10000
LUT_ARGB_255 = [None] * 0x10000
LUT_BGRA_255 = [None] * 0x10000
LUT_ABGR_255 = [None] * 0x10000


def unpack(color):
    """
    Return the color as a quadruple (r, g, b, a), where a is 0 or 1
    (0 meaning opaque and 1 meaning transparent), and r, g and b are
    between 0 and 31 inclusive.
    """
    return (color & 0x1F,
            (color >> 5) & 0x1F,
            (color >> 10) & 0x1F,
            (color >> 15) & 1)


def pack(r, g, b, a=0):
    """
    Pack these channel values into a color. a should be 0 or 1 (0
    meaning opaque and 1 meaning transparent), and r, g and b should be
    between 0 and 31 inclusive.
    """
    value = 0
    value |= r & 0x1F
    value |= (g & 0x1F) << 5
    value |= (b & 0x1F) << 10
    value |= (a & 1) << 15
    return value


def unpack255(color):
    """
    Same as unpack(), but scales each channel to the standard 0-255
    scale.
    """
    r, g, b, a = unpack(color)
    r = r << 3 | r >> 2
    g = g << 3 | g >> 2
    b = b << 3 | b >> 2
    return (r, g, b, 0 if a else 255)


def pack255(r, g, b, a=255):
    """
    Same as pack(), but expects each channel to be on the standard 0-255
    scale.
    """
    return pack(((r + 4) << 2) // 33,
                ((g + 4) << 2) // 33,
                ((b + 4) << 2) // 33,
                1 if a < 128 else 0)


# Avoiding using the already-defined functions here, because doing so
# provides a significant speedup
for i in range(0x10000):
    r = i & 0x1F
    g = (i >> 5) & 0x1F
    b = (i >> 10) & 0x1F
    a = (i >> 15) & 1
    LUT_UNPACKED[i] = (r, g, b, a)

    r255 = r << 3 | r >> 2
    g255 = g << 3 | g >> 2
    b255 = b << 3 | b >> 2
    a255 = 0 if a else 255
    LUT_UNPACKED_255[i] = (r255, g255, b255, a255)

    LUT_RGBA_255[i] = (r255 << 24) | (g255 << 16) | (b255 << 8) | a255
    LUT_ARGB_255[i] = (a255 << 24) | (r255 << 16) | (g255 << 8) | b255
    LUT_BGRA_255[i] = (b255 << 24) | (g255 << 16) | (r255 << 8) | a255
    LUT_ABGR_255[i] = (a255 << 24) | (b255 << 16) | (g255 << 8) | r255



# Copyright 2019 RoadrunnerWMC
#
# This file is part of ndspy.
#
# ndspy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# ndspy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with ndspy.  If not, see <https://www.gnu.org/licenses/>.

import enum
import struct

from . import _common


class TextureFormat(enum.IntEnum):
    """
    An enum describing the various texture formats (for 3D models) the
    Nintendo DS supports.
    """
    UNKNOWN_0 = 0
    TRANSLUCENT_A3I5 = 1
    PALETTED_2BPP = 2
    PALETTED_4BPP = 3
    PALETTED_8BPP = 4
    TEXELED_4X4 = 5
    TRANSLUCENT_A5I3 = 6
    DIRECT_16_BIT = 7


    def bitsPerPixel1(self):
        """
        Return the number of bits per pixel that this format requires in
        a texture's first data region.
        This is useful for calculating the expected amount of data some
        texture will have, given its width, height and format.
        """
        return {
            type(self).UNKNOWN_0: 0,
            type(self).TRANSLUCENT_A3I5: 8,
            type(self).PALETTED_2BPP: 2,
            type(self).PALETTED_4BPP: 4,
            type(self).PALETTED_8BPP: 8,
            type(self).TEXELED_4X4: 2,
            type(self).TRANSLUCENT_A5I3: 8,
            type(self).DIRECT_16_BIT: 16,
            }[self]


    def bitsPerPixel2(self):
        """
        Return the number of bits per pixel that this format requires in
        a texture's second data region.
        This is useful for calculating the expected amount of data some
        texture will have, given its width, height and format.
        """
        # Only TEXELED_4X4 uses the second data region, and it's 1bpp
        # there. 
        if self == type(self).TEXELED_4X4:
            return 1
        else:
            return 0


class TextureCoordinatesTransformationMode(enum.IntEnum):
    """
    An enum describing the four ways texture coordinates can be
    transformed.
    See https://problemkaputt.de/gbatek.htm#ds3dtexturecoordinates
    for more.
    """
    NONE = 0
    TEX_COORD = 1
    NORMAL = 2
    VERTEX = 3


class Texture:
    """
    A texture for a 3D model. May depend on a palette.
    """
    def __init__(self, unk00, unk02, params, unk04, data1, data2):
        self.unk00 = unk00
        self.unk02 = unk02

        self.repeatS = bool(params & 1)                 # TODO: figure out what this means
        self.repeatT = bool(params & 2)                 # TODO: figure out what this means
        self.mirrorS = bool(params & 4)                 # TODO: figure out what this means
        self.mirrorT = bool(params & 8)                 # TODO: figure out what this means
        self.width = 8 << ((params >> 4) & 7)
        self.height = 8 << ((params >> 7) & 7)
        self.format = TextureFormat((params >> 10) & 7)
        self.isColor0Transparent = bool(params & 0x2000)
        self.coordsTransformationMode = TextureCoordinatesTransformationMode(
            (params >> 14) & 3)

        self.unk04 = unk04
        self.data1 = data1
        self.data2 = data2


    @classmethod
    def fromFlags(cls, unk00, unk02, repeatS, repeatT, mirrorS, mirrorT,
            width, height, format, isColor0Transparent,
            coordsTransformationMode, unk04, data1, data2):
        self = cls(unk00, unk02, 0, unk04, data1, data2)
        self.repeatS = repeatS
        self.repeatT = repeatT
        self.mirrorS = mirrorS
        self.mirrorT = mirrorT
        self.width = width
        self.height = height
        self.format = format
        self.isColor0Transparent = isColor0Transparent
        self.coordsTransformationMode = coordsTransformationMode
        return self


    def save(self):
        params = 0

        if self.repeatS: params |= 1
        if self.repeatT: params |= 2
        if self.mirrorS: params |= 4
        if self.mirrorT: params |= 8

        SIZES_ENCODED = {
            8:    0,
            16:   1,
            32:   2,
            64:   3,
            128:  4,
            256:  5,
            512:  6,
            1024: 7,
        }

        if self.width not in SIZES_ENCODED:
            raise ValueError(f'Texture width (currently {self.width})'
                ' must be a power of 2 between 8 and 1024 inclusive!')
        if self.height not in SIZES_ENCODED:
            raise ValueError(f'Texture height (currently {self.height})'
                ' must be a power of 2 between 8 and 1024 inclusive!')

        params |= SIZES_ENCODED[self.width] << 4
        params |= SIZES_ENCODED[self.height] << 7

        params |= (self.format & 7) << 10
        if self.isColor0Transparent: params |= 0x2000
        params |= (self.coordsTransformationMode & 3) << 14

        return self.unk00, self.unk02, params, self.unk04, self.data1, self.data2


    def __str__(self):
        format = {
            int(TextureFormat.UNKNOWN_0): 'unknown-0',
            int(TextureFormat.TRANSLUCENT_A3I5): 'translucent-a3i5',
            int(TextureFormat.PALETTED_2BPP): 'paletted-2bpp',
            int(TextureFormat.PALETTED_4BPP): 'paletted-4bpp',
            int(TextureFormat.PALETTED_8BPP): 'paletted-8bpp',
            int(TextureFormat.TEXELED_4X4): 'texeled-4x4',
            int(TextureFormat.TRANSLUCENT_A5I3): 'translucent-a5i3',
            int(TextureFormat.DIRECT_16_BIT): 'direct-16-bit',
            }[self.format]
        return f'<texture {format} {self.width}x{self.height}>'


    def __repr__(self):
        try:
            argsList = list(self.save())

            # Params
            argsList[2] = hex(argsList[2])

            # Datas 1 & 2
            argsList[4] = _common.shortBytesRepr(argsList[4])
            argsList[5] = _common.shortBytesRepr(argsList[5])

            args = ', '.join(str(x) for x in argsList)

        except Exception:
            args = '...'

        return f'{type(self).__name__}({args})'


class Palette:
    """
    A palette as used in 3D model textures.
    Contains a colors list as well as some other metadata.
    """
    def __init__(self, unk00, unk02, unkHeader02, data):
        self.unk00 = unk00
        self.unk02 = unk02
        self.unkHeader02 = unkHeader02

        if len(data) % 2 == 1:
            raise ValueError(f'Palette data length is {len(data)} (must be even)')
        self.colors = list(struct.unpack(f'<{len(data) // 2}H', data))


    def save(self):
        data = struct.pack(f'<{len(self.colors)}H', *self.colors)
        return self.unk00, self.unk02, self.unkHeader02, data


    @classmethod
    def fromColors(cls, unk00, unk02, unkHeader02, colors):
        self = cls(unk00, unk02, unkHeader02, b'')
        self.colors = colors
        return self


    def __str__(self):
        return f'<palette ({len(self.colors)} colors) {_common.shortColorsListRepr(self.colors)}>'


    def __repr__(self):
        return (f'{type(self).__name__}.fromColors('
            f'{self.unk00},'
            f' {self.unk02},'
            f' {self.unkHeader02},'
            f' {_common.shortColorsListRepr(self.colors)})')


class NSBTX:
    """
    Similar to NSBMD, but only contains textures and palettes (no 3D
    data).
    """
    def __init__(self, data=None):
        if data is not None:
            self._initFromData(data)


    def _initFromData(self, data):
        """
        Load the NSBTX from file data.
        """
        if not data.startswith(b'BTX0'):
            raise ValueError('Invalid NSBTX: incorrect magic')

        magic, bom, version, filesize, headersize, numblocks = \
            _common.NDS_STD_FILE_HEADER.unpack_from(data, 0)
        if version != 1:
            raise ValueError(f'Unsupported NSBTX version: {version}')
        
        assert numblocks == 1

        TEX0 = data[0x14:]
        assert TEX0.startswith(b'TEX0')

        parsed = _readTEX0(TEX0)
        self.unk08 = parsed['unk08']
        self.unk10 = parsed['unk10']
        self.unk18 = parsed['unk18']
        self.unk20 = parsed['unk20']
        self.unk2C = parsed['unk2C']
        self.unk32 = parsed['unk32']
        self.textures = parsed['textures']
        self.palettes = parsed['palettes']


    @classmethod
    def fromTexturesAndPalettes(cls, unk08, unk10, unk18, unk20, unk2C, unk32, textures, palettes):
        self = cls()
        self.unk08 = unk08
        self.unk10 = unk10
        self.unk18 = unk18
        self.unk20 = unk20
        self.unk2C = unk2C
        self.unk32 = unk32
        self.textures = textures
        self.palettes = palettes
        return self


    def save(self):
        """
        Save the NSBTX back to a file.
        """
        texInfo = {}
        texInfo['unk08'] = self.unk08
        texInfo['unk10'] = self.unk10
        texInfo['unk18'] = self.unk18
        texInfo['unk20'] = self.unk20
        texInfo['unk2C'] = self.unk2C
        texInfo['unk32'] = self.unk32
        texInfo['textures'] = self.textures
        texInfo['palettes'] = self.palettes
        TEX0 = _saveTEX0(texInfo)

        data = bytearray(0x14)
        data.extend(TEX0)

        # Add the NDS standard file header
        _common.NDS_STD_FILE_HEADER.pack_into(data, 0,
            b'BTX0', 0xFEFF, 1, len(data), 0x10, 1)

        # Insert the offset to the TEX0 block
        data[0x10] = 0x14

        # And return.
        return bytes(data)


def _readTEX0(data):
    """
    Read a TEX0 block.
    Return a bunch of data as a dict.
    """
    assert data.startswith(b'TEX0')

    returnVal = {}

    # TEX0 header
    (unk08, texDataLen, texOff,
            unk10, texDataOff, unk18, compressedData1Len, compressedInfoOff,
            unk20, compressedData1Off, compressedData2Off, unk2C,
            palDataLen, unk32, palOff, palDataOff) = \
        struct.unpack_from('<IHH3IHH4IHH2I', data, 0x8)
    returnVal['unk08'] = unk08
    returnVal['unk10'] = unk10
    returnVal['unk18'] = unk18
    returnVal['unk20'] = unk20
    returnVal['unk2C'] = unk2C
    returnVal['unk32'] = unk32

    texDataLen <<= 3
    palDataLen <<= 3
    compressedData1Len <<= 3
    compressedData2Len = compressedData1Len // 2

    # Textures
    returnVal['textures'] = []
    for name, unk00, unk02, texHeader in _common.loadInfoBlock(data, texOff, 8):
        thisTexOff, params, unk04 = struct.unpack('<HHI', texHeader)
        thisTexOff <<= 3

        tex = Texture(unk00, unk02, params, unk04, b'', b'')

        if tex.format == TextureFormat.TEXELED_4X4:
            thisTexOff1 = compressedData1Off + texDataOff
            thisTexOff2 = compressedData2Off + texDataOff // 2
            thisTexDataLen1 = int(tex.width * tex.height * tex.format.bitsPerPixel1() / 8)
            thisTexDataLen2 = int(tex.width * tex.height * tex.format.bitsPerPixel2() / 8)
            thisTexData1 = data[thisTexOff1 : thisTexOff1 + thisTexDataLen1]
            thisTexData2 = data[thisTexOff2 : thisTexOff2 + thisTexDataLen2]
        else:
            thisTexOff += texDataOff
            thisTexDataLen = int(tex.width * tex.height * tex.format.bitsPerPixel1() / 8)
            thisTexData1 = data[thisTexOff : thisTexOff+thisTexDataLen]
            thisTexData2 = b''

        tex.data1 = thisTexData1
        tex.data2 = thisTexData2

        returnVal['textures'].append((name, tex))

    # Palettes
    # This needs to be done in 2 passes because palettes don't have a
    # "data length" value: first we collect all the palette data start
    # positions (and metadata), then we sort the data start positions,
    # and then we make the actual Palette objects with the palette data.
    paletteStartOffsets = []
    palette2StartOffset = {}

    paletteTuples = []
    returnVal['palettes'] = []

    for name, unk00, unk02, palHeader in _common.loadInfoBlock(data, palOff, 4):
        thisPalOff, unkHeader02 = struct.unpack('<HH', palHeader)
        thisPalOff <<= 3
        thisPalOff += palDataOff

        pt = (name, unk00, unk02, unkHeader02)

        paletteStartOffsets.append(thisPalOff)
        palette2StartOffset[pt] = thisPalOff
        paletteTuples.append(pt)

    paletteStartOffsets.append(palDataOff + palDataLen)
    paletteStartOffsets.sort()

    for pt in paletteTuples:
        startOff = palette2StartOffset[pt]
        endOff = paletteStartOffsets[paletteStartOffsets.index(startOff) + 1]
        palData = data[startOff : endOff]

        name, unk00, unk02, unkHeader02 = pt
        pal = Palette(unk00, unk02, unkHeader02, palData)
        returnVal['palettes'].append((name, pal))

    return returnVal


def _saveTEX0(fields):
    """
    Given a dict containing TEX0 data (in the format returned by
    _loadTEX0), return a bytes object representing the TEX0 block.
    """

    data = bytearray(0x3C)
    texturesData = bytearray()
    compressedData1 = bytearray()
    compressedData2 = bytearray()
    palettesData = bytearray()

    # Textures header
    compressedInfoOff = len(data)
    texOff = len(data)
    
    entries = []
    for texName, tex in fields['textures']:
        unk00, unk02, params, unk04, data1, data2 = tex.save()

        if tex.format == TextureFormat.TEXELED_4X4:
            thisTexOff = len(compressedData1)
            if len(data1) != len(data2) * 2:
                raise ValueError(f'For texeled 4x4 textures, data1 must'
                    ' be twice as long as data2!'
                    f' (Found lengths: {len(data1)}, {len(data2)})')
            compressedData1.extend(data1)
            compressedData2.extend(data2)
        else:
            thisTexOff = len(texturesData)
            texturesData.extend(data1)

        entryData = struct.pack('<HHI', thisTexOff >> 3, params, unk04)
        entries.append((texName, unk00, unk02, entryData))

    data.extend(_common.saveInfoBlock(entries, 8))

    # Palettes header
    palOff = len(data)
    
    entries = []
    for palName, pal in fields['palettes']:
        unk00, unk02, unkHeader02, thisPaletteData = pal.save()
        if len(thisPaletteData) % 8:
            # 4 colors = 8 bytes
            raise ValueError('Palette must be a multiple of 4 colors long')
        thisPalOff = len(palettesData)
        palettesData.extend(thisPaletteData)

        entryData = struct.pack('<HH', thisPalOff >> 3, unkHeader02)
        entries.append((palName, unk00, unk02, entryData))

    data.extend(_common.saveInfoBlock(entries, 4))

    # Data bytearrays
    texDataOff = len(data)
    data.extend(texturesData)
    compressedData1Off = len(data)
    data.extend(compressedData1)
    compressedData2Off = len(data)
    data.extend(compressedData2)
    palDataOff = len(data)
    data.extend(palettesData)

    # TEX0 header
    struct.pack_into('<4sIIHH3IHH4IHH2I', data, 0,
        b'TEX0', len(data), fields['unk08'], len(texturesData) >> 3, texOff,
        fields['unk10'], texDataOff, fields['unk18'], len(compressedData1) >> 3, compressedInfoOff,
        fields['unk20'], compressedData1Off, compressedData2Off, fields['unk2C'],
        len(palettesData) >> 3, fields['unk32'], palOff, palDataOff)

    return bytes(data)



# Copyright 2019 RoadrunnerWMC
#
# This file is part of ndspy.
#
# ndspy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# ndspy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with ndspy.  If not, see <https://www.gnu.org/licenses/>.
"""
Support for LZ10 compression.
"""


import struct

from . import _lzCommon


def decompress(data):
    """
    Decompress LZ10-compressed data.
    """

    # NOTE:
    # This code is ported from NSMBe, which was converted from Elitemap.

    if data[0] != 0x10:
        raise TypeError("This isn't a LZ10-compressed file.")

    dataLen = struct.unpack_from('<I', data)[0] >> 8

    out = bytearray(dataLen)
    inPos, outPos = 4, 0

    while dataLen > 0:
        d = data[inPos]; inPos += 1

        if d:
            for i in range(8):
                if d & 0x80:
                    thing, = struct.unpack_from('>H', data, inPos); inPos += 2

                    length = (thing >> 12) + 3
                    offset = thing & 0xFFF
                    windowOffset = outPos - offset - 1

                    for j in range(length):
                        out[outPos] = out[windowOffset]
                        outPos += 1; windowOffset += 1; dataLen -= 1

                        if dataLen == 0:
                            return bytes(out)

                else:
                    out[outPos] = data[inPos]
                    outPos += 1; inPos += 1; dataLen -= 1

                    if dataLen == 0:
                        return bytes(out)

                d <<= 1
        else:
            for i in range(8):
                out[outPos] = data[inPos]
                outPos += 1; inPos += 1; dataLen -= 1

                if dataLen == 0:
                    return bytes(out)

    return bytes(out)


def decompressFromFile(filePath):
    """
    Load a LZ10-compressed filesystem file, and decompress it.
    """
    with open(filePath, 'rb') as f:
        return decompress(f.read())


def compress(data):
    """
    Compress data in LZ10 format.
    """

    # NOTE:
    # This code is ported from NSMBe.

    compressed, _, _ = _lzCommon.compress(data, 1, 0x1000, 18, True, False)
    compressed = bytearray(compressed)
    compressed[:0] = struct.pack('<I', (len(data) << 8) | 0x10)
    return bytes(compressed)


def compressToFile(data, filePath):
    """
    Compress data in LZ10 format, and save it to a filesystem file.
    """
    d = compress(data)
    with open(filePath, 'wb') as f:
        f.write(d)



# Copyright 2019 RoadrunnerWMC
#
# This file is part of ndspy.
#
# ndspy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# ndspy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with ndspy.  If not, see <https://www.gnu.org/licenses/>.
"""
Support for SWAV sound waves.
"""

import enum
import struct

from . import WaveType
from . import _common


class SWAV:
    """
    An SWAV streamed audio file.
    """

    def __init__(self, file=None):
        self.waveType = WaveType.PCM8
        self.isLooped = False
        self.sampleRate = 8000
        self.time = 0
        self.loopOffset = 0
        self.totalLength = 0
        self.data = b''

        if file is not None:
            if not file.startswith(b'SWAV'):
                raise ValueError("Wrong magic (should be b'SWAV', instead "
                                 f'found {file[:4]})')

            self._initFromData(file)


    def _initFromData(self, file):
        """
        Initialize the SWAV from file data.
        """

        magic, bom, version, filesize, headersize, numblocks = \
            _common.NDS_STD_FILE_HEADER.unpack_from(file, 0)
        if version != 0x100:
            raise ValueError(f'Unsupported SWAV version: {version}')
        assert magic == b'SWAV'

        dataMagic, dataSize = struct.unpack_from('<4sI', file, 0x10)
        assert dataMagic == b'DATA'

        (waveType, self.isLooped, self.sampleRate, self.time, self.loopOffset,
            loopLength) = struct.unpack_from('<B?3HI', file, 0x18)
        self.waveType = WaveType(waveType)
        self.totalLength = self.loopOffset + loopLength

        self.data = file[0x24:filesize]


    @classmethod
    def fromData(cls, data=b'', *, waveType=WaveType.PCM8, isLooped=False,
            sampleRate=8000, time=0, loopOffset=0, totalLength=0):
        """
        Create an SWAV from raw waveform data.
        """
        self = cls()
        self.waveType = waveType
        self.isLooped = isLooped
        self.sampleRate = sampleRate
        self.time = time
        self.loopOffset = loopOffset
        self.totalLength = totalLength
        self.data = data
        return self


    @classmethod
    def fromFile(cls, filePath):
        """
        Load an SWAV from a filesystem file.
        """
        with open(filePath, 'rb') as f:
            return cls(f.read())


    def save(self, *, updateTime=False, updateTotalLength=False):
        """
        Generate file data representing this SWAV.
        """

        if updateTime:
            self.time = int(16756991 / self.sampleRate)

        if updateTotalLength:
            self.totalLength = len(self.data) // 4

        if self.loopOffset > self.totalLength:
            raise ValueError(f'Loop offset ({self.loopOffset}) > total length ({self.totalLength})!')

        data = bytearray()
        data.extend(_common.NDS_STD_FILE_HEADER.pack(
            b'SWAV', 0xFEFF, 0x100, 0x24 + len(self.data), 0x10, 1))
        data.extend(struct.pack(
            '<4sI', b'DATA', 0x14 + len(self.data)))
        data.extend(struct.pack(
            '<B?3HI', self.waveType, self.isLooped, self.sampleRate, self.time,
            self.loopOffset, self.totalLength - self.loopOffset))
        data.extend(self.data)
        return bytes(data)


    def saveToFile(self, filePath, *, updateTime=False, updateTotalLength=False):
        """
        Generate file data representing this SWAV, and save it to a
        filesystem file.
        """
        d = self.save(updateTime=updateTime,
                      updateTotalLength=updateTotalLength)
        with open(filePath, 'wb') as f:
            f.write(d)


    def __str__(self):
        try:
            waveType = WaveType(self.waveType).name
        except Exception:
            waveType = f'waveType={hex(self.waveType)}'

        looped = ' looped' if self.isLooped else ''

        return f'<swav {waveType} {self.sampleRate}Hz{looped}>'


    def __repr__(self):
        args = []
        args.append(_common.shortBytesRepr(self.data))
        args.append(f'waveType={self.waveType!r}')
        args.append(f'isLooped={self.isLooped!r}')
        args.append(f'sampleRate={self.sampleRate!r}')
        args.append(f'time={self.time!r}')
        args.append(f'loopOffset={self.loopOffset!r}')
        args.append(f'totalLength={self.totalLength!r}')
        return f'{type(self).__name__}.fromData({", ".join(args)})'



import setuptools

with open('README.md', 'r', encoding='utf-8') as fh:
    long_description = fh.read()

setuptools.setup(
    name='ndspy',
    version='1.0.1',
    author='RoadrunnerWMC',
    author_email='roadrunnerwmc@gmail.com',
    description='Python library that can help you read, modify and create many types of files used in Nintendo DS games.',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/RoadrunnerWMC/ndspy',
    packages=setuptools.find_packages(),
    python_requires='>=3.6',
    install_requires=[
        'crcmod',
    ],
    classifiers=[
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)',
        'Operating System :: OS Independent',
    ],
)


# Copyright 2019 RoadrunnerWMC
#
# This file is part of ndspy.
#
# ndspy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# ndspy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with ndspy.  If not, see <https://www.gnu.org/licenses/>.


def compress(data, posSubtract, maxMatchDiff, maxMatchLen, zerosAtEnd,
        searchReverse):
    """
    A configurable function suitable for all LZ-like algorithms.
    Returns the compressed data, the amount of data at the end that
    can be left uncompressed (if the format supports that, such as code
    compression), and the amount of compressed data that that data
    comprises.

    This code is primarily ported from NSMBe.
    """

    def compressionSearch(pos):
        """
        Find the longest match in `data` (nonlocal) at or after `pos`.
        This function has been rewritten in place of NSMBe's,
        to optimize its performance in Python.
        (A straight port of NSMBe's algorithm caused some files to take
        over 40 seconds to compress. With this version, all files I've
        tested take less than one second, and the compressed files
        match the old algorithm's output byte for byte.)
        """

        start = max(0, pos - maxMatchDiff)

        # Strategy: do a binary search of potential match sizes, to
        # find the longest match that exists in the data.

        lower = 0
        upper = min(maxMatchLen, len(data) - pos)

        recordMatchPos = recordMatchLen = 0
        while lower <= upper:
            # Attempt to find a match at the middle length
            matchLen = (lower + upper) // 2
            match = data[pos : pos + matchLen]
            if searchReverse:
                matchPos = data.rfind(match, start, pos)
            else:
                matchPos = data.find(match, start, pos)

            if matchPos == -1:
                # No such match -- any matches will be smaller than this
                upper = matchLen - 1
            else:
                # Match found!
                if matchLen > recordMatchLen:
                    recordMatchPos, recordMatchLen = matchPos, matchLen
                lower = matchLen + 1

        return recordMatchPos, recordMatchLen

    result = bytearray()

    current = 0 # Index of current byte to compress

    ignorableDataAmount = 0
    ignorableCompressedAmount = 0

    bestSavingsSoFar = 0

    while current < len(data):
        blockFlags = 0

        # We'll go back and fill in blockFlags at the end of the loop.
        blockFlagsOffset = len(result)
        result.append(0)
        ignorableCompressedAmount += 1

        for i in range(8):

            # Not sure if this is needed. The DS probably ignores this data.
            if current >= len(data):
                if zerosAtEnd:
                    result.append(0)
                continue

            searchPos, searchLen = compressionSearch(current)
            searchDisp = current - searchPos - posSubtract

            if searchLen > 2:
                # We found a big match; let's write a compressed block
                blockFlags |= 1 << (7 - i)

                result.append((((searchLen - 3) & 0xF) << 4) | ((searchDisp >> 8) & 0xF))
                result.append(searchDisp & 0xFF)
                current += searchLen

                ignorableDataAmount += searchLen
                ignorableCompressedAmount += 2

            else:
                result.append(data[current])
                current += 1
                ignorableDataAmount += 1
                ignorableCompressedAmount += 1

            savingsNow = current - len(result)
            if savingsNow > bestSavingsSoFar:
                ignorableDataAmount = 0
                ignorableCompressedAmount = 0
                bestSavingsSoFar = savingsNow

        result[blockFlagsOffset] = blockFlags

    return bytes(result), ignorableDataAmount, ignorableCompressedAmount


# Copyright 2019 RoadrunnerWMC
#
# This file is part of ndspy.
#
# ndspy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# ndspy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with ndspy.  If not, see <https://www.gnu.org/licenses/>.


import struct

from . import _common
from . import color


HavePIL = True
try:
    import PIL.Image
except ImportError:
    HavePIL = False


def loadPalette(data):
    """
    Convert binary data to a list of (r, g, b, a) color tuples.
    This is the inverse of savePalette().
    """
    colors = []
    colorVals = struct.unpack_from(f'<{len(data) // 2}H', data)
    colors = [color.LUT_UNPACKED_255[c] for c in colorVals]
    return colors


def savePalette(colors):
    """
    Convert a list of (r, g, b, a) color tuples to binary data.
    This is the inverse of loadPalette().
    """
    colorVals = [color.pack255(r, g, b, a) for (r, g, b, a) in colors]
    return struct.pack(f'<{len(colors)}H', *colorVals)


class ImageTile:
    """
    A class that represents a single image tile.
    """
    data = None
    is4bpp = False

    def __init__(self, data, is4bpp=False):
        """
        data should be bytes or bytearray
        """
        self.data = data
        self.is4bpp = is4bpp


    def render(self, colors, paletteNum=0):
        """
        Given a list of colors and the desired palette number to use, 
        render this ImageTile as a PIL Image.
        colors should be a list of (r, g, b, a) tuples.
        """
        if not HavePIL:
            raise RuntimeError('PIL is not installed, so ndspy cannot render ImageTiles.')

        cs = paletteNum * (16 if self.is4bpp else 256)
        img = PIL.Image.new('RGBA', (8, 8), (0, 0, 0, 0))
        for y in range(8):
            if self.is4bpp:
                for x in range(0, 8, 2):
                    col12 = self.data[y * 4 + x // 2]
                    col1, col2 = col12 & 0xF, col12 >> 4
                    if col1:
                        img.putpixel((x, y), colors[cs + col1])
                    if col2:
                        img.putpixel((x+1, y), colors[cs + col2])
            else:
                for x in range(8):
                    col = self.data[y * 8 + x]
                    if col:
                        img.putpixel((x, y), colors[cs + col])

        return img


def loadImageTiles(data, is4bpp=False):
    """
    Convert binary data to a list of ImageTiles.
    This is the inverse of saveImageTiles().
    """
    bytesPerTile = 32 if is4bpp else 64
    tileCount = len(data) // bytesPerTile

    tiles = []
    for i in range(tileCount):
        start = i * bytesPerTile
        tiles.append(ImageTile(data[start : start+bytesPerTile], is4bpp))

    return tiles


def saveImageTiles(tiles):
    """
    Convert a list of ImageTiles to binary data.
    This is the inverse of loadImageTiles().
    """
    return b''.join(t.data for t in tiles)


class TilemapTile:
    """
    A class representing a single tilemap tile
    """
    tileNum = 0
    hFlip = False
    vFlip = False
    paletteNum = 0

    def __init__(self, value=0):
        """
        Initialize the TilemapTile from a raw 16-bit data value.
        """
        self.value = value


    @classmethod
    def fromParameters(cls, tileNum, hFlip=False, vFlip=False, paletteNum=0):
        """
        Create a new TilemapTile with the parameters given.
        """
        self = cls()
        self.tileNum = tileNum
        self.hFlip = hFlip
        self.vFlip = vFlip
        self.paletteNum = paletteNum
        return self


    @property
    def value(self):
        """
        This tile's 16-bit data value.
        """
        v = 0
        v |= self.tileNum & 0x3FF
        if self.hFlip: v |= 0x400
        if self.vFlip: v |= 0x800
        v |= (self.paletteNum & 0xF) << 12
        return v

    @value.setter
    def value(self, v):
        self.tileNum = v & 0x3FF
        self.hFlip = bool(v & 0x400)
        self.vFlip = bool(v & 0x800)
        self.paletteNum = (v >> 12) & 0xF


    def renderSingle(self, imageTile, colors):
        """
        Given an image tile and a list of colors, render this
        TilemapTile as a PIL Image.
        imageTile should be an ImageTile.
        colors should be a list of (r, g, b, a) tuples.
        """
        if not HavePIL:
            raise RuntimeError('PIL is not installed, so ndspy cannot render TilemapTiles.')

        img = imageTile.render(colors, tile.paletteNum)
        if self.hFlip:
            img = img.transpose(PIL.Image.FLIP_LEFT_RIGHT)
        if self.vFlip:
            img = img.transpose(PIL.Image.FLIP_TOP_BOTTOM)
        return img


    def render(self, imageTiles, colors, tileNumOffset=0):
        """
        Given lists of image tiles and colors, render this TilemapTile
        as a PIL Image.
        imageTiles should be a list of, well, ImageTiles.
        colors should be a list of (r, g, b, a) tuples.
        """
        if self.tileNum < tileNumOffset:
            raise ValueError(f'TilemapTile.tileNum < tileNumOffset ({self.tileNum} < {tileNumOffset})')
        return self.renderSingle(imageTiles[self.tileNum - tileNumOffset],
                                 colors)


def loadTilemapTiles(data):
    """
    Convert binary data to a list of TilemapTiles.
    This is the inverse of saveTilemapTiles().
    """
    tiles = []
    for i in range(0, len(data), 2):
        t, = struct.unpack_from('<H', data, i)
        tiles.append(TilemapTile(t))
    return tiles


def saveTilemapTiles(tiles):
    """
    Convert a list of TilemapTiles to binary data.
    This is the inverse of loadTilemapTiles().
    """
    return b''.join(struct.pack('<H', t.value) for t in tiles)


def renderImageTiles(tiles, colors, paletteNum=0, width=32):
    """
    Given a list of ImageTiles, a list of colors, a desired palette
    ID to render with, render all the tiles as a single image.
    The width of the image defaults to 32 tiles, but can be adjusted.
    """
    if not HavePIL:
        raise RuntimeError('PIL is not installed, so ndspy cannot render ImageTiles.')

    height, remainder = divmod(len(tileImages), width)
    if remainder: height += 1

    img = PIL.Image.new('RGBA', (8 * width, 8 * height), (0, 0, 0, 0))

    tileIter = iter(tiles)
    for y in range(height):
        for x in range(width):
            tileImg = next(tileIter).render(colors, paletteNum)
            img.paste(tileImg, (x * 8, y * 8))

    return img


def renderTilemap(tilemapTiles, imageTiles, colors, width, tileNumOffset=0):
    """
    Given a list of TilemapTiles, a list of ImageTiles, a list of
    colors, render the tilemap to an image. The tilemap's width must
    also be specified.
    """
    height, remainder = divmod(len(tilemapTiles), width)
    if remainder: height += 1

    img = PIL.Image.new('RGBA', (8 * width, 8 * height), (0, 0, 0, 0))

    tilemapIter = iter(tilemapTiles)
    for y in range(height):
        for x in range(width):
            tileImg = next(tilemapIter).render(imageTiles, colors, tileNumOffset)
            img.paste(tileImg, (x * 8, y * 8))

    return img


def tileAt(tiles, x, y, w=32):
    """
    Convenience function: return the tile at (x, y) in the tile list
    given, assuming a row width of `w` (32 by default).
    """
    return tiles[y * w + x]


def putTile(tiles, x, y, t, w=32):
    """
    Convenience function: replace the tile at (x, y) in the tile list
    given with `t`, assuming a row width of `w` (32 by default).
    If (x, y) is not in the tile list, nothing happens.
    """
    if x >= w or y >= w: return
    if x < 0 or y < 0: return
    try:
        tiles[y * w + x] = t
    except IndexError: pass
